import itertools
import logging
from pathlib import Path
from typing import List

import numpy as np
import pandas as pd

from immuneML.data_model.datasets.RepertoireDataset import RepertoireDataset
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.encoding_reports.EncodingReport import EncodingReport
from immuneML.util.PathBuilder import PathBuilder


class Matches(EncodingReport):
    """
    Reports the number of matches that were found when using one of the following encoders:

    * :ref:`MatchedSequences` encoder
    * :ref:`MatchedReceptors` encoder
    * :ref:`MatchedRegex` encoder


    Report results are:

    * A table containing all matches, where the rows correspond to the Repertoires, and the
      columns correspond to the objects to match (regular expressions or receptor sequences).
    * The repertoire sizes (read frequencies and the number of unique sequences per repertoire), for each of the chains.
      This can be used to calculate the percentage of matched sequences in a repertoire.
    * When using :ref:`MatchedSequences` encoder or
      :ref:`MatchedReceptors` encoder, tables describing
      the chains and receptors (ids, chains, V and J genes and sequences).
    * When using :ref:`MatchedReceptors` encoder or using
      :ref:`MatchedRegex` encoder with chain pairs, tables describing
      the paired matches (where a match was found in both chains) per repertoire.


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            reports:
                my_match_report: Matches
    """

    @classmethod
    def build_object(cls, **kwargs):
        return Matches(**kwargs)

    def __init__(self, dataset: RepertoireDataset = None, result_path: Path = None, name: str = None, number_of_processes: int = 1):
        super().__init__(dataset=dataset, result_path=result_path, name=name, number_of_processes=number_of_processes)

    def _generate(self) -> ReportResult:
        PathBuilder.build(self.result_path)
        return self._write_reports()

    def _write_reports(self) -> ReportResult:
        all_matches_table = self._write_match_table()
        repertoire_sizes = self._write_repertoire_sizes()

        output_tables = [all_matches_table, repertoire_sizes]

        if self.dataset.encoded_data.encoding == "MatchedSequencesEncoder":
            output_tables += self._write_sequence_info(self.result_path / "sequence_info")
        else:
            if len(self.dataset.encoded_data.feature_annotations["locus"].unique()) == 2:
                output_tables += self._write_paired_matches(self.result_path / "paired_matches")

            if self.dataset.encoded_data.encoding == "MatchedReceptorsEncoder":
                output_tables += self._write_receptor_info(self.result_path / "receptor_info")

        return ReportResult(self.name,
                            info=f"Reports the number of matches that were found when using {self.dataset.encoded_data.encoding}",
                            output_tables=output_tables)

    def _write_match_table(self):
        id_df = pd.DataFrame({"repertoire_id": self.dataset.encoded_data.example_ids,
                              'subject_id': self.dataset.get_metadata(['subject_id'])['subject_id']})
        label_df = pd.DataFrame(self.dataset.encoded_data.labels)
        matches_df = pd.DataFrame(self.dataset.encoded_data.examples, columns=self.dataset.encoded_data.feature_names)
        different_cols = label_df.columns.difference(id_df.columns)

        result_path = self.result_path / "complete_match_count_table.csv"
        id_df.join(label_df[different_cols]).join(matches_df).to_csv(result_path, index=False)

        return ReportOutput(result_path, "All matches")

    def _write_paired_matches(self, paired_matches_path: Path) -> List[ReportOutput]:
        PathBuilder.build(paired_matches_path)

        report_outputs = []
        for i in range(0, len(self.dataset.encoded_data.example_ids)):
            file_name = "example_{}_".format(self.dataset.encoded_data.example_ids[i])
            file_name += "_".join(["{label}_{value}".format(label=label, value=values[i]) for
                                  label, values in self.dataset.encoded_data.labels.items()])
            file_name += ".csv"
            file_path = paired_matches_path / file_name

            if self.dataset.encoded_data.encoding == "MatchedReceptorsEncoder":
                self._write_paired_receptor_matches_for_repertoire(self.dataset.encoded_data.examples[i], file_path)
            elif self.dataset.encoded_data.encoding == "MatchedRegexEncoder":
                self._write_paired_regex_matches_for_repertoire(self.dataset.encoded_data.examples[i], file_path)

            report_outputs.append(ReportOutput(file_path, f"Example {self.dataset.encoded_data.example_ids[i]} paired matches"))

        return report_outputs

    def _write_paired_receptor_matches_for_repertoire(self, matches, filename):
        match_identifiers = []
        match_values = []

        for i in range(0, int(len(matches) / 2)):
            first_match_idx = i * 2
            second_match_idx = i * 2 + 1

            if matches[first_match_idx] > 0 and matches[second_match_idx] > 0:
                match_identifiers.append(self.dataset.encoded_data.feature_names[first_match_idx])
                match_identifiers.append(self.dataset.encoded_data.feature_names[second_match_idx])
                match_values.append(matches[first_match_idx])
                match_values.append(matches[second_match_idx])

        results_df = pd.DataFrame([match_values], columns=match_identifiers)
        results_df.to_csv(filename, index=False)

    def _write_paired_regex_matches_for_repertoire(self, matches, filename):
        match_identifiers = []
        match_values = []

        annotation_df = self.dataset.encoded_data.feature_annotations

        for cell_id in sorted(set(annotation_df["regex_id"])):
            chain_ids = list(annotation_df.loc[annotation_df["regex_id"] == cell_id]["locus_id"])

            if len(chain_ids) == 2:
                first_match_idx = self.dataset.encoded_data.feature_names.index(chain_ids[0])
                second_match_idx = self.dataset.encoded_data.feature_names.index(chain_ids[1])

                if matches[first_match_idx] > 0 and matches[second_match_idx] > 0:
                    match_identifiers.append(chain_ids[0])
                    match_identifiers.append(chain_ids[1])
                    match_values.append(matches[first_match_idx])
                    match_values.append(matches[second_match_idx])

        results_df = pd.DataFrame([match_values], columns=match_identifiers)
        results_df.to_csv(filename, index=False)

    def _write_repertoire_sizes(self):
        """
        Writes the repertoire sizes (# clones & # reads) per subject, per chain.
        """
        all_subjects = sorted(set(self.dataset.get_metadata(['subject_id'])['subject_id']))
        all_chains = sorted(set(self.dataset.encoded_data.feature_annotations["locus"]))

        results_df = pd.DataFrame(list(itertools.product(all_subjects, all_chains)),
                                  columns=["subject_id", "locus"])
        results_df["n_reads"] = 0
        results_df["n_clones"] = 0

        for repertoire in self.dataset.repertoires:
            rep_counts = repertoire.data.duplicate_count
            rep_chains = repertoire.data.locus.tolist()

            for chain in all_chains:
                indices = [el == chain for el in rep_chains]
                results_df.loc[(results_df.subject_id == repertoire.metadata["subject_id"]) & (results_df.locus == chain),
                               'n_reads'] += np.sum(rep_counts[indices])
                results_df.loc[(results_df.subject_id == repertoire.metadata["subject_id"]) & (results_df.locus == chain),
                               'n_clones'] += len(rep_counts[indices])

        results_path = self.result_path / "repertoire_sizes.csv"
        results_df.to_csv(results_path, index=False)

        return ReportOutput(results_path, "Repertoire sizes")

    def _write_receptor_info(self, receptor_info_path) -> List[ReportOutput]:
        PathBuilder.build(receptor_info_path)

        receptor_chains = self.dataset.encoded_data.feature_annotations
        chain_types = receptor_chains["locus"].unique()

        first_chains = receptor_chains.loc[receptor_chains.locus == chain_types[0]]
        second_chains = receptor_chains.loc[receptor_chains.locus == chain_types[1]]

        first_chains.drop(columns=["locus"], inplace=True)
        second_chains.drop(columns=["locus"], inplace=True)

        on_cols = ["cell_id"]
        if "clonotype_id" in second_chains.columns and first_chains.columns:
            on_cols += ["clonotype_id"]

        receptors = pd.merge(first_chains, second_chains,
                             on=on_cols,
                             suffixes=(f"_{chain_types[0]}", f"_{chain_types[1]}"))

        unique_alpha_chains = first_chains.drop_duplicates(subset=["sequence", "v_call", "j_call"])
        unique_beta_chains = second_chains.drop_duplicates(subset=["sequence", "v_call", "j_call"])
        unique_receptors = receptors.drop_duplicates(subset=[f"sequence_{chain_types[0]}", f"v_call_{chain_types[0]}", f"j_call_{chain_types[0]}",
                                                             f"sequence_{chain_types[1]}", f"v_call_{chain_types[1]}", f"j_call_{chain_types[1]}"])

        receptor_chains_path = receptor_info_path / "all_chains.csv"
        receptor_chains.to_csv(receptor_chains_path, index=False)
        receptors_path = receptor_info_path / "all_receptors.csv"
        receptors.to_csv(receptors_path, index=False)
        unique_chain1_path = receptor_info_path / f"unique_{chain_types[0]}_chains.csv"
        unique_alpha_chains.to_csv(unique_chain1_path, index=False)
        unique_chain2_path = receptor_info_path / f"unique_{chain_types[1]}_chains.csv"
        unique_beta_chains.to_csv(unique_chain2_path, index=False)
        unique_receptors_path = receptor_info_path / "unique_receptors.csv"
        unique_receptors.to_csv(unique_receptors_path, index=False)

        return [ReportOutput(path=path, name=name) for path, name in [(receptors_path, "All receptors info"),
                                                                      (receptor_chains_path, "All receptor chains info"),
                                                                      (unique_receptors_path, "Unique receptors info"),
                                                                      (unique_chain1_path, "Unique chain 1 info"),
                                                                      (unique_chain2_path, "Unique chain 2 info")]]

    def _write_sequence_info(self, sequence_info_path) -> List[ReportOutput]:
        PathBuilder.build(sequence_info_path)

        chains = self.dataset.encoded_data.feature_annotations
        unique_chains = chains.drop_duplicates(subset=["sequence", "v_call", "j_call"])

        chains_path = sequence_info_path / "all_chains.csv"
        chains.to_csv(chains_path, index=False)
        unique_chains_path = sequence_info_path / "unique_chains.csv"
        unique_chains.to_csv(unique_chains_path, index=False)

        return [ReportOutput(path=path, name=name) for path, name in [(chains_path, "All chains info"), (unique_chains_path, "Unique chains info")]]

    def check_prerequisites(self):
        if self.dataset.encoded_data is None or self.dataset.encoded_data.examples is None:
            logging.warning(f"No encoding was specified for dataset {self.dataset.identifier}. Please use one of the following encodings: MatchedReceptorsEncoder, MatchedSequencesEncoder, MatchedRegexEncoder. Matches report will not be created.")
            return False

        if self.dataset.encoded_data.encoding not in ("MatchedReceptorsEncoder", "MatchedSequencesEncoder", "MatchedRegexEncoder"):
            logging.warning(f"Encoding {self.dataset.encoded_data.encoding} is not compatible with this report type. Matches report will not be created.")
            return False
        else:
            return True



from pathlib import Path

import numpy as np
import pandas as pd

from immuneML.analysis.criteria_matches.CriteriaMatcher import CriteriaMatcher
from immuneML.analysis.criteria_matches.CriteriaTypeInstantiator import CriteriaTypeInstantiator
from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.data_model.datasets.RepertoireDataset import RepertoireDataset
from immuneML.preprocessing.filters.Filter import Filter
from immuneML.util.PathBuilder import PathBuilder


class MetadataFilter(Filter):
    """
    Removes examples from a dataset based on the examples' metadata. It works for any dataset type. Note that
    for repertoire datasets, this means that repertoires will be filtered out, and for sequences datasets - sequences.

    Since this filter changes the number of examples, it cannot be used with
    :ref:`TrainMLModel` instruction. Use with DatasetExport instruction instead.

    **Specification arguments:**

    - criteria (dict): a nested dictionary that specifies the criteria for keeping the dataset examples based on the
      column values; it contains the type of evaluation, name of the column, and additional parameters depending on
      evaluation; alternatively, it can contain a combination of multiple (evaluation, column, parameters) groups;
      evaluation_types: IN, NOT_IN, NOT_NA, GREATER_THAN, LESS_THAN, TOP_N, RANDOM_N; for IN, NOT_IN the parameter name
      is 'values', for GREATER_THAN, LESS_THAN the parameter name is 'threshold' and for TOP_N, RANDOM_N the parameter
      name is 'number'; supported boolean combinations of groups are AND and OR with (evaluation, column, parameter)
      groups specified under 'operands' key; see the YAML below for example.


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        preprocessing_sequences:
            my_preprocessing:
                - my_filter:
                    # Example filter that keeps e.g., repertoires with values greater than 1 in the "my_column_name"
                    # column of the metadata_file
                    MetadataFilter:
                        type: GREATER_THAN
                        column: my_column_name
                        threshold: 1
            my_second_preprocessing:
                - my_filter2: # only examples which in column "label" have values 'label_val1' or 'label_val2' are kept
                    MetadataFilter:
                        type: IN
                        values: [label_val1, label_val2]
                        column: label
            my_third_preprocessing_example:
                - my_combined_filter:
                    MetadataFilter:
                    # keeps examples with that have label_val1 or label_val2 in the column label and
                    # that at the same time have a value larger than 1.3 in another_metadata_column
                        type: AND
                        operands:
                        - type: IN
                          values: [label_val1, label_val2]
                          column: label
                        - type: GREATER_THAN
                          column: another_metadata_column
                          threshold: 1.3

    """

    def __init__(self, criteria: dict, result_path: Path = None):
        super().__init__(result_path)
        self.criteria = CriteriaTypeInstantiator.instantiate(criteria)

    def keeps_example_count(self) -> bool:
        return False

    def process_dataset(self, dataset: Dataset, result_path: Path, number_of_processes=1):
        self.result_path = result_path if result_path is not None else self.result_path
        PathBuilder.build(self.result_path)
        indices = self._get_matching_indices(dataset)
        return dataset.make_subset(indices, self.result_path, "filtered")

    def _get_matching_indices(self, dataset: Dataset):
        metadata = dataset.get_metadata(None, True)
        matches = CriteriaMatcher().match(self.criteria, metadata)
        indices = np.where(matches)[0]
        return indices





import copy
from pathlib import Path
from typing import List

from immuneML.data_model.SequenceParams import RegionType
from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.environment.Constants import Constants
from immuneML.environment.Label import Label
from immuneML.environment.LabelConfiguration import LabelConfiguration
from immuneML.environment.SequenceType import SequenceType
from immuneML.example_weighting.ExampleWeightingParams import ExampleWeightingParams
from immuneML.example_weighting.ExampleWeightingStrategy import ExampleWeightingStrategy
from immuneML.hyperparameter_optimization.HPSetting import HPSetting
from immuneML.hyperparameter_optimization.config.SplitConfig import SplitConfig
from immuneML.hyperparameter_optimization.states.HPSelectionState import HPSelectionState
from immuneML.hyperparameter_optimization.states.TrainMLModelState import TrainMLModelState
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.ReportUtil import ReportUtil
from immuneML.util.Logger import print_log
from immuneML.util.PathBuilder import PathBuilder
from immuneML.workflows.steps.DataWeighter import DataWeighter
from immuneML.workflows.steps.DataEncoder import DataEncoder
from immuneML.workflows.steps.DataEncoderParams import DataEncoderParams
from immuneML.workflows.steps.DataWeighterParams import DataWeighterParams
from immuneML.workflows.steps.MLMethodAssessment import MLMethodAssessment
from immuneML.workflows.steps.MLMethodAssessmentParams import MLMethodAssessmentParams
from immuneML.workflows.steps.data_splitter.DataSplitter import DataSplitter
from immuneML.workflows.steps.data_splitter.DataSplitterParams import DataSplitterParams


class HPUtil:

    @staticmethod
    def split_data(dataset: Dataset, split_config: SplitConfig, path: Path, label_config: LabelConfiguration) -> tuple:
        paths = [path / f"split_{i + 1}" for i in range(split_config.split_count)]
        params = DataSplitterParams(
            dataset=dataset,
            split_strategy=split_config.split_strategy,
            split_count=split_config.split_count,
            training_percentage=split_config.training_percentage,
            paths=paths,
            split_config=split_config,
            label_config=label_config
        )
        return DataSplitter.run(params)

    @staticmethod
    def get_average_performance(performances):
        if performances is not None and isinstance(performances, list) and len(performances) > 0 and all(
                isinstance(perf, float) for perf in performances):
            return sum(perf for perf in performances) / len(performances)
        else:
            return Constants.NOT_COMPUTED

    @staticmethod
    def preprocess_dataset(dataset: Dataset, preproc_sequence: list, path: Path, context: dict = None,
                           hp_setting: HPSetting = None, number_of_processes: int = 1) -> Dataset:
        if dataset is not None:
            if isinstance(preproc_sequence, list) and len(preproc_sequence) > 0:
                PathBuilder.build(path)
                tmp_dataset = dataset.clone() if context is None or "dataset" not in context else context["dataset"]

                for preprocessing in preproc_sequence:
                    tmp_dataset = preprocessing.process_dataset(tmp_dataset, path, number_of_processes)
                    print_log(f"Preprocessed {tmp_dataset.__class__.__name__.split('Dataset')[0].lower()} dataset {tmp_dataset.name} with {preprocessing.__class__.__name__}:\n"
                        f"- Example count: {tmp_dataset.get_example_count()}\n"
                        f"- Labels: {tmp_dataset.get_label_names()}", True)

                if context is not None and "dataset" in context:
                    context["preprocessed_dataset"] = {str(hp_setting): tmp_dataset}
                    indices = [i for i in range(context["dataset"].get_example_count())
                               if context["dataset"].repertoires[i].identifier in dataset.get_example_ids()]
                    preprocessed_dataset = tmp_dataset.make_subset(indices, path, Dataset.PREPROCESSED)
                else:
                    preprocessed_dataset = tmp_dataset

                return preprocessed_dataset
            else:
                return dataset

    @staticmethod
    def weight_examples(dataset, weighting_strategy: ExampleWeightingStrategy, path: Path, learn_model: bool,
                        number_of_processes: int):
        weighted_dataset = DataWeighter.run(DataWeighterParams(
            dataset=dataset,
            weighting_strategy=weighting_strategy,
            weighting_params=ExampleWeightingParams(
                result_path=path,
                pool_size=number_of_processes,
                learn_model=learn_model
            ),
        ))
        return weighted_dataset

    @staticmethod
    def encode_dataset(dataset, hp_setting: HPSetting, path: Path, learn_model: bool, context: dict,
                       number_of_processes: int,
                       label_configuration: LabelConfiguration, encode_labels: bool = True,
                       sequence_type: SequenceType = SequenceType.AMINO_ACID,
                       region_type: RegionType = RegionType.IMGT_CDR3):
        PathBuilder.build(path)

        encoded_dataset = DataEncoder.run(DataEncoderParams(
            dataset=dataset,
            encoder=hp_setting.encoder,
            encoder_params=EncoderParams(
                model=hp_setting.encoder_params,
                result_path=path,
                pool_size=number_of_processes,
                label_config=label_configuration,
                learn_model=learn_model,
                encode_labels=encode_labels,
                sequence_type=sequence_type,
                region_type=region_type
            ),
        ))
        return encoded_dataset

    @staticmethod
    def assess_performance(method, metrics, optimization_metric, dataset, split_index, current_path: Path,
                           test_predictions_path: Path, label: Label, ml_score_path: Path):
        return MLMethodAssessment.run(MLMethodAssessmentParams(
            method=method,
            dataset=dataset,
            predictions_path=test_predictions_path,
            split_index=split_index,
            label=label,
            metrics=metrics,
            optimization_metric=optimization_metric,
            path=current_path,
            ml_score_path=ml_score_path
        ))

    @staticmethod
    def run_hyperparameter_reports(state: TrainMLModelState, path: Path) -> List[ReportResult]:
        report_results = []
        for key, report in state.reports.items():
            tmp_report = copy.deepcopy(report)
            tmp_report.state = state
            tmp_report.result_path = path / key
            tmp_report.number_of_processes = state.number_of_processes
            report_result = tmp_report.generate_report()
            report_results.append(report_result)
        return report_results

    @staticmethod
    def run_selection_reports(state: TrainMLModelState, dataset, train_datasets: list, val_datasets: list,
                              selection_state: HPSelectionState):
        path = selection_state.path
        data_split_reports = state.selection.reports.data_split_reports.values()
        for index in range(len(train_datasets)):
            split_reports_path = path / f"split_{index + 1}"

            selection_state.train_data_reports += ReportUtil.run_data_reports(train_datasets[index], data_split_reports,
                                                                              split_reports_path / "data_reports_train",
                                                                              state.number_of_processes, state.context)
            selection_state.val_data_reports += ReportUtil.run_data_reports(val_datasets[index], data_split_reports,
                                                                            split_reports_path / "data_reports_test",
                                                                            state.number_of_processes, state.context)

        data_reports = state.selection.reports.data_reports.values()
        selection_state.data_reports = ReportUtil.run_data_reports(dataset, data_reports, path / "reports",
                                                                   state.number_of_processes, state.context)



from sklearn.neighbors import KNeighborsClassifier

from immuneML.ml_methods.classifiers.SklearnMethod import SklearnMethod
from scripts.specification_util import update_docs_per_mapping


class KNN(SklearnMethod):
    """
    This is a wrapper of scikit-learn’s KNeighborsClassifier class.
    This ML method creates a distance matrix using the given encoded data. If the encoded data is already a distance
    matrix (for example, when using the :ref:`Distance` or :ref:`CompAIRRDistance` encoders), please use :ref:`PrecomputedKNN` instead.

    Please see the `KNeighborsClassifier scikit-learn documentation <https://scikit-learn.org/stable/modules/generated/sklearn.neighbors.KNeighborsClassifier.html>`_
    of KNeighborsClassifier for the parameters.

    For usage instructions, check :py:obj:`~immuneML.ml_methods.classifiers.SklearnMethod.SklearnMethod`.

    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            ml_methods:
                my_knn_method:
                    KNN:
                        # sklearn parameters (same names as in original sklearn class)
                        weights: uniform # always use this setting for weights
                        n_neighbors: [5, 10, 15] # find the optimal number of neighbors
                        # Additional parameter that determines whether to print convergence warnings
                        show_warnings: True
                    # if any of the parameters under KNN is a list and model_selection_cv is True,
                    # a grid search will be done over the given parameters, using the number of folds specified in model_selection_n_folds,
                    # and the optimal model will be selected
                    model_selection_cv: True
                    model_selection_n_folds: 5
                # alternative way to define ML method with default values:
                my_default_knn: KNN

    """

    def __init__(self, parameter_grid: dict = None, parameters: dict = None):
        parameters = parameters if parameters is not None else {}
        parameter_grid = parameter_grid if parameter_grid is not None else {}

        super(KNN, self).__init__(parameter_grid=parameter_grid, parameters=parameters)

    def _get_ml_model(self, cores_for_training: int = 2, X=None):
        params = self._parameters
        params["n_jobs"] = cores_for_training
        return KNeighborsClassifier(**params)

    def get_params(self):
        return self.model.get_params(deep=True)

    def can_predict_proba(self) -> bool:
        return True

    def can_fit_with_example_weights(self) -> bool:
        return False

    def get_compatible_encoders(self):
        from immuneML.encodings.distance_encoding.DistanceEncoder import DistanceEncoder
        from immuneML.encodings.evenness_profile.EvennessProfileEncoder import EvennessProfileEncoder
        from immuneML.encodings.abundance_encoding.SequenceAbundanceEncoder import SequenceAbundanceEncoder
        from immuneML.encodings.abundance_encoding.CompAIRRSequenceAbundanceEncoder import CompAIRRSequenceAbundanceEncoder
        from immuneML.encodings.abundance_encoding.KmerAbundanceEncoder import KmerAbundanceEncoder
        from immuneML.encodings.kmer_frequency.KmerFrequencyEncoder import KmerFrequencyEncoder
        from immuneML.encodings.onehot.OneHotEncoder import OneHotEncoder
        from immuneML.encodings.word2vec.Word2VecEncoder import Word2VecEncoder
        return [KmerFrequencyEncoder, OneHotEncoder, Word2VecEncoder, SequenceAbundanceEncoder,
                CompAIRRSequenceAbundanceEncoder, KmerAbundanceEncoder, EvennessProfileEncoder, DistanceEncoder]

    @staticmethod
    def get_documentation():
        doc = str(KNN.__doc__)

        mapping = {
            "For usage instructions, check :py:obj:`~immuneML.ml_methods.classifiers.SklearnMethod.SklearnMethod`.": SklearnMethod.get_usage_documentation("KNN"),
        }

        doc = update_docs_per_mapping(doc, mapping)
        return doc



from abc import ABC
from pathlib import Path

from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.ml_methods.clustering.ClusteringMethod import ClusteringMethod
from immuneML.reports.Report import Report


class ClusteringMethodReport(Report, ABC):
    """
    Clustering method reports show some features or statistics about the clustering method.
    """

    DOCS_TITLE = "Clustering method reports"

    def __init__(self, dataset: Dataset = None, model: ClusteringMethod = None, result_path: Path = None,
                 name: str = None):
        super().__init__(name=name, result_path=result_path)
        self.dataset = dataset
        self.model = model



import logging
import warnings
from pathlib import Path

import numpy as np
import pandas as pd

from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.hyperparameter_optimization.HPSetting import HPSetting
from immuneML.ml_methods.classifiers.LogisticRegression import LogisticRegression
from immuneML.ml_methods.classifiers.MLMethod import MLMethod
from immuneML.ml_methods.classifiers.RandomForestClassifier import RandomForestClassifier
from immuneML.ml_methods.classifiers.SVC import SVC
from immuneML.ml_methods.classifiers.SVM import SVM
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.ml_reports.MLReport import MLReport
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder


class MotifSeedRecovery(MLReport):
    """
    This report can be used to show how well implanted motifs (for example, through the Simulation instruction) can
    be recovered by various machine learning methods using the k-mer encoding.
    This report creates a boxplot, where the x axis (box grouping) represents the maximum possible overlap between
    an implanted motif seed and a kmer feature (measured in number of positions), and the y axis shows the coefficient size
    of the respective kmer feature. If the machine learning method has learned the implanted motif seeds, the coefficient
    size is expected to be largest for the kmer features with high overlap to the motif seeds.

    Note that to use this report, the following criteria must be met:

    - KmerFrequencyEncoder must be used.
    - One of the following classifiers must be used: RandomForestClassifier, LogisticRegression, SVM, SVC
    - For each label, the implanted motif seeds relevant to that label must be specified

    To find the overlap score between kmer features and implanted motif seeds, the two sequences are compared in a sliding
    window approach, and the maximum overlap is calculated.

    Overlap scores between kmer features and implanted motifs are calculated differently based on the Hamming distance that was
    allowed during implanting.

    .. indent with spaces
    .. code-block:: text

        Without hamming distance:
        Seed:     AAA  -> score = 3
        Feature: xAAAx
                  ^^^

        Seed:     AAA  -> score = 0
        Feature: xAAxx

        With hamming distance:
        Seed:     AAA  -> score = 3
        Feature: xAAAx
                  ^^^

        Seed:     AAA  -> score = 2
        Feature: xAAxx
                  ^^

        Furthermore, gap positions in the motif seed are ignored:
        Seed:     A/AA  -> score = 3
        Feature: xAxAAx
                  ^/^^

    See :ref:`Recovering simulated immune signals` for more details.


    Example output:

    .. image:: ../../_static/images/reports/motif_seed_recovery.png
       :alt: Motif seed recovery report
       :width: 650


    **Specification arguments:**

    - implanted_motifs_per_label (dict): a nested dictionary that specifies the motif seeds that were implanted in the given dataset. The first
      level of keys in this dictionary represents the different labels. In the inner dictionary there should be two keys: "seeds" and
      "hamming_distance":

      - seeds: a list of motif seeds. The seeds may contain gaps, specified by a '/' symbol.

      - hamming_distance: A boolean value that specifies whether hamming distance was allowed when implanting the motif
        seeds for a given label. Note that this applies to all seeds for this label.

      - gap_sizes: a list of all the possible gap sizes that were used when implanting a gapped motif seed. When no
        gapped seeds are used, this value has no effect.


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            reports:
                my_motif_report:
                    MotifSeedRecovery:
                        implanted_motifs_per_label:
                            CD:
                                seeds:
                                - AA/A
                                - AAA
                                hamming_distance: False
                                gap_sizes:
                                - 0
                                - 1
                                - 2
                            T1D:
                                seeds:
                                - CC/C
                                - CCC
                                hamming_distance: True
                                gap_sizes:
                                - 2

    """

    @classmethod
    def build_object(cls, **kwargs):
        ParameterValidator.assert_keys_present(list(kwargs.keys()),
                                               ["implanted_motifs_per_label"],
                                               "MotifSeedRecovery", "MotifSeedRecovery report")

        implanted_motifs_per_label = kwargs["implanted_motifs_per_label"]

        ParameterValidator.assert_type_and_value(implanted_motifs_per_label, dict,
                                                 "MotifSeedRecovery",
                                                 f"implanted_motifs_per_label")

        for label_name in implanted_motifs_per_label.keys():
            ParameterValidator.assert_type_and_value(implanted_motifs_per_label[label_name], dict,
                                                     "MotifSeedRecovery",
                                                     f"implanted_motifs_per_label/{label_name}")

            ParameterValidator.assert_keys_present(implanted_motifs_per_label[label_name].keys(),
                                                   ["hamming_distance", "seeds", "gap_sizes"],
                                                   "MotifSeedRecovery", f"implanted_motifs_per_label/{label_name}")
            ParameterValidator.assert_type_and_value(implanted_motifs_per_label[label_name]["hamming_distance"], bool,
                                                     "MotifSeedRecovery", f"implanted_motifs_per_label/{label_name}/hamming_distance")
            ParameterValidator.assert_type_and_value(implanted_motifs_per_label[label_name]["gap_sizes"], list,
                                                     "MotifSeedRecovery",
                                                     f"implanted_motifs_per_label/{label_name}/gap_sizes")
            ParameterValidator.assert_type_and_value(implanted_motifs_per_label[label_name]["seeds"], list,
                                                     "MotifSeedRecovery", f"implanted_motifs_per_label/{label_name}/seeds")
            for gap_size in implanted_motifs_per_label[label_name]["gap_sizes"]:
                ParameterValidator.assert_type_and_value(gap_size, int, "MotifSeedRecovery",
                                                         f"implanted_motifs_per_label/{label_name}/gap_sizes", min_inclusive=0)
            for seed in implanted_motifs_per_label[label_name]["seeds"]:
                ParameterValidator.assert_type_and_value(seed, str, "MotifSeedRecovery",
                                                         f"implanted_motifs_per_label/{label_name}/seeds")

        return MotifSeedRecovery(implanted_motifs_per_label)

    def __init__(self, implanted_motifs_per_label, train_dataset: Dataset = None, test_dataset: Dataset = None,
                 method: MLMethod = None, result_path: Path = None, name: str = None, hp_setting: HPSetting = None,
                 label=None, number_of_processes: int = 1):
        super().__init__(train_dataset=train_dataset, test_dataset=test_dataset, method=method, result_path=result_path,
                         name=name, hp_setting=hp_setting, label=label, number_of_processes=number_of_processes)

        self.implanted_motifs_per_label = implanted_motifs_per_label
        self._param_field = None
        self._y_axis_title = None
        self._x_axis_title = None

    def _generate(self):
        PathBuilder.build(self.result_path)

        self._set_plotting_parameters()

        plot_df = self._retrieve_plot_data()
        report_output_table = self._write_results_table(plot_df)
        report_output_fig = self._plot(plot_df, "motif_seed_recovery")

        return ReportResult(self.name,
                            info="This report shows how well implanted ('ground truth') motifs are recovered by ML models using the k-mer encoding. The x axis (box grouping) represents the maximum number of overlapping positions between a 'ground truth' motif seed and a k-mer feature. The y axis values represent the learned coefficients. ",
                            output_tables=[report_output_table],
                            output_figures=[report_output_fig])

    def _write_results_table(self, plotting_data):
        filepath = self.result_path / "motif_seed_recovery.csv"
        plotting_data.to_csv(filepath, index=False)
        return ReportOutput(path=filepath, name="motif seed recovery csv")

    def _set_plotting_parameters(self):
        if isinstance(self.method, RandomForestClassifier):
            self._param_field = "feature_importances"
            self._y_axis_title = "Feature importance"
        else:
            # SVM, logistic regression, ...
            self._param_field = "coefficients"
            self._y_axis_title = "Coefficient value"

        if self.implanted_motifs_per_label[self.label.name]["hamming_distance"]:
            self._x_axis_title = "Positions overlap between feature and motif seeds<br>(hamming distance allowed)"
        else:
            self._x_axis_title = "Positions overlap between feature and motif seeds"

    def _retrieve_plot_data(self):
        seeds = self._get_implanted_seeds()
        overlap_fn = self._get_overlap_fn()
        features = self._retrieve_feature_names()

        plot_df = self.calculate_seed_overlap(seeds, features, overlap_fn)
        plot_df["coefficients"] = self.method.get_params()[self._param_field]
        return plot_df

    def _get_implanted_seeds(self):
        return self.implanted_motifs_per_label[self.label.name]["seeds"]

    def _get_overlap_fn(self):
        is_hamming_distance = self.implanted_motifs_per_label[self.label.name]["hamming_distance"]
        overlap_fn = self.hamming_overlap if is_hamming_distance else self.identical_overlap
        return overlap_fn

    def _retrieve_feature_names(self):
        if self.train_dataset and self.train_dataset.encoded_data:
            return self.train_dataset.encoded_data.feature_names

    def _plot(self, plotting_data, output_name):
        if plotting_data.empty:
            logging.warning(f"Coefficients: empty data subset specified, skipping {output_name} plot...")
        else:

            filename = self.result_path / f"{output_name}.html"

            import plotly.express as px
            figure = px.box(plotting_data, x="max_seed_overlap", y="coefficients", labels={
                "max_seed_overlap": self._x_axis_title,
                "coefficients": self._y_axis_title
            }, template='plotly_white', color_discrete_sequence=px.colors.diverging.Tealrose)

            figure.write_html(str(filename))

            return ReportOutput(filename, f"Overlap between implanted motif seeds and features versus {self._y_axis_title.lower()}")

    def hamming_overlap(self, seed, feature):
        return sum(np.array(list(seed)) == np.array(list(feature)))

    def identical_overlap(self, seed, feature):
        if "/" in seed:
            exclude_idx_start = seed.index("/")
            exclude_idx_end = seed.rindex("/")
            seed = seed[:exclude_idx_start] + seed[exclude_idx_end + 1:]
            feature = feature[:exclude_idx_start] + feature[exclude_idx_end + 1:]

        while feature.startswith("-"):
            feature = feature[1:]
            seed = seed[1:]

        while feature.endswith("-"):
            feature = feature[:-1]
            seed = seed[:-1]

        return int(seed == feature) * len(seed)

    def max_overlap_sliding(self, seed, feature, overlap_fn):
        max_score = 0

        sizes = self.implanted_motifs_per_label[self.label.name]["gap_sizes"]

        for gap_size in sizes:
            gap_adjusted_seed = seed.replace("/", "/" * gap_size)

            padding = "-" * (len(gap_adjusted_seed) - 1)

            padded_feature = padding + feature + padding

            for start_idx in range(0, len(feature) + len(padding)):
                feature_slice = padded_feature[start_idx:start_idx + len(gap_adjusted_seed)]
                max_score = max(max_score, overlap_fn(gap_adjusted_seed, feature_slice))

        return max_score

    def calculate_seed_overlap(self, motif_seeds, features, overlap_fn):
        seed_df = pd.DataFrame({"features": features})

        for seed in motif_seeds:
            seed_df[seed] = [self.max_overlap_sliding(seed, feature, overlap_fn) for feature in seed_df["features"]]

        seed_df["max_seed_overlap"] = seed_df.drop("features", axis=1).max(axis=1)
        seed_df = seed_df[["features", "max_seed_overlap"]]

        return seed_df

    def check_prerequisites(self):
        location = "MotifSeedRecovery"

        run_report = True

        if not any([isinstance(self.method, legal_method) for legal_method in (RandomForestClassifier, LogisticRegression, SVM, SVC)]):
            logging.warning(f"{location} report can only be created for RandomForestClassifier, LogisticRegression, SVC, or SVM, but got "
                            f"{type(self.method).__name__} instead. Report {self.name} will not be created.")
            run_report = False

        if self.label.name not in self.implanted_motifs_per_label.keys():
            logging.warning(
                f"{location}: no implanted motifs were specified for the label '{self.label}'. "
                f"These motifs should be specified under 'implanted_motifs_per_label'. Report {self.name} will not be created.")
            run_report = False

        if self.train_dataset.encoded_data is None or self.train_dataset.encoded_data.examples is None or self.train_dataset.encoded_data.feature_names is None:
            logging.warning(
                f"{location}: this report can only be created for an encoded dataset with specified feature names. Report {self.name} will not be created.")
            run_report = False

        if self.train_dataset.encoded_data.encoding != "KmerFrequencyEncoder":
            logging.warning(
                f"{location}: this report can only be created for a dataset encoded with the KmerFrequencyEncoder. Report {self.name} will not be created.")
            run_report = False

        return run_report



from abc import ABC
from pathlib import Path

from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.ml_methods.generative_models.GenerativeModel import GenerativeModel
from immuneML.reports.Report import Report


class GenModelReport(Report, ABC):
    """
    Generative model reports show some type of features or statistics about a generative model.
    """

    DOCS_TITLE = "Generative model reports"

    def __init__(self, dataset: Dataset = None, model: GenerativeModel = None, result_path: Path = None,
                 name: str = None):
        super().__init__(name=name, result_path=result_path)
        self.dataset = dataset
        self.model = model



import abc
import os
import shutil
from abc import ABCMeta
from pathlib import Path
import glob
import traceback
from immuneML.presentation.html.FailedGalaxyHTMLBuilder import FailedGalaxyHTMLBuilder


class GalaxyTool(metaclass=ABCMeta):

    def __init__(self, specification_path: Path, result_path: Path, **kwargs):
        self.yaml_path = Path(specification_path) if specification_path is not None else None
        self.result_path = Path(os.path.relpath(result_path)) if result_path is not None else None

    def run(self):
        try:
            self._run()
        except Exception as e:
            traceback_str = traceback.format_exc()
            self._make_failed_galaxy_run_html(traceback_str)
            raise e
        finally:
            shutil.make_archive(Path("./immuneML_output"), "zip", self.result_path)
            shutil.move(str(Path("./immuneML_output.zip")), str(self.result_path))

    @abc.abstractmethod
    def _run(self):
        pass

    def _make_failed_galaxy_run_html(self, traceback_str):
        FailedGalaxyHTMLBuilder.build(self.result_path, traceback_str)




from dataclasses import fields as get_fields
from itertools import chain

import numpy as np
from npstructures import RaggedArray


def pad_ragged_array(new_array, target_shape, padded_value):
    """pad ragged array to match sequence lengths"""
    padded_array = RaggedArray([[padded_value for _ in range(target_shape[1][ind])] for ind in range(target_shape[0])])
    for row_ind in range(target_shape[0]):
        np.put(padded_array[row_ind], 0, new_array[row_ind])

    return padded_array


def merge_dataclass_objects(objects: list):  # TODO: replace with equivalent from npstructures
    field_names = sorted(list(set(chain.from_iterable([field.name for field in get_fields(obj)] for obj in objects))))

    for obj in objects:
        assert all(hasattr(obj, field) for field in field_names), (obj, field_names)

    cls = type(objects[0])
    return cls(**{field_name: list(chain.from_iterable([getattr(obj, field_name) for obj in objects])) for field_name in
                  field_names})



import abc

from immuneML.data_model.datasets.Dataset import Dataset


class ClusteringMethod:
    """

    Clustering methods are algorithms which can be used to cluster repertoires, receptors or
    sequences without using external label information (such as disease or antigen binding state)

    These methods can be used in the :ref:`Clustering` instruction.

    """
    DOCS_TITLE = "Clustering methods"

    def __init__(self, name: str = None):
        self.name = name
        self.model = None

    @abc.abstractmethod
    def fit(self, dataset: Dataset):
        pass

    @abc.abstractmethod
    def predict(self, dataset: Dataset):
        pass

    @abc.abstractmethod
    def fit_predict(self, dataset: Dataset):
        pass

    @abc.abstractmethod
    def transform(self, dataset: Dataset):
        pass


def get_data_for_clustering(dataset: Dataset):
    if dataset.encoded_data.dimensionality_reduced_data is not None:
        return dataset.encoded_data.dimensionality_reduced_data
    else:
        return dataset.encoded_data.examples



from sklearn.svm import SVC

from immuneML.ml_methods.classifiers.SklearnMethod import SklearnMethod
from scripts.specification_util import update_docs_per_mapping


class SVM(SklearnMethod):
    """
    This is a wrapper of scikit-learn’s SVC class. Please see the
    `SVC scikit-learn documentation <https://scikit-learn.org/stable/modules/generated/sklearn.svm.SVC.html>`_
    of SVC for the parameters.

    Note: if you are interested in plotting the coefficients of the SVM model,
    consider running the :ref:`Coefficients` report.

    For usage instructions, check :py:obj:`~immuneML.ml_methods.classifiers.SklearnMethod.SklearnMethod`.


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            ml_methods:
                my_svm: # user-defined method name
                    SVM: # name of the ML method
                        # sklearn parameters (same names as in original sklearn class)
                        C: [0.01, 0.1, 1, 10, 100] # find the optimal value for C
                        kernel: linear
                        # Additional parameter that determines whether to print convergence warnings
                        show_warnings: True
                    # if any of the parameters under SVM is a list and model_selection_cv is True,
                    # a grid search will be done over the given parameters, using the number of folds specified in model_selection_n_folds,
                    # and the optimal model will be selected
                    model_selection_cv: True
                    model_selection_n_folds: 5
                # alternative way to define ML method with default values:
                my_default_svm: SVM

    """

    def __init__(self, parameter_grid: dict = None, parameters: dict = None):
        _parameter_grid = parameter_grid if parameter_grid is not None else {}
        _parameters = parameters if parameters is not None else {}

        super(SVM, self).__init__(parameter_grid=_parameter_grid, parameters=_parameters)

    def _get_ml_model(self, cores_for_training: int = 2, X=None):
        return SVC(**self._parameters)

    def can_predict_proba(self) -> bool:
        return False

    def can_fit_with_example_weights(self) -> bool:
        return False

    def get_params(self):
        params = self.model.get_params()
        params["coefficients"] = self.model.coef_[0].tolist() if self.model.kernel == 'linear' else self.model.dual_coef_.toarray().tolist() if hasattr(self.model.dual_coef_, "toarray") else self.model.dual_coef_.tolist()
        params["intercept"] = self.model.intercept_.tolist()
        return params

    @staticmethod
    def get_documentation():
        doc = str(SVM.__doc__)

        mapping = {
            "For usage instructions, check :py:obj:`~immuneML.ml_methods.classifiers.SklearnMethod.SklearnMethod`.": SklearnMethod.get_usage_documentation("SVM"),
        }

        doc = update_docs_per_mapping(doc, mapping)
        return doc



from enum import Enum


class SymbolType(Enum):

    MOTIF = 0
    SIGNAL = 1
    ML_METHOD = 2
    DATASET = 3
    ENCODING = 4
    REPORT = 5
    SIMULATION = 6
    PREPROCESSING = 7
    INSTRUCTION = 8
    OUTPUT = 9
    WEIGHTING = 10



from pathlib import Path

from immuneML.environment.Label import Label
from immuneML.environment.LabelConfiguration import LabelConfiguration
from immuneML.hyperparameter_optimization.HPSetting import HPSetting
from immuneML.hyperparameter_optimization.config.SplitType import SplitType
from immuneML.hyperparameter_optimization.core.HPUtil import HPUtil
from immuneML.hyperparameter_optimization.states.HPSelectionState import HPSelectionState
from immuneML.hyperparameter_optimization.states.TrainMLModelState import TrainMLModelState
from immuneML.util.Logger import print_log
from immuneML.util.PathBuilder import PathBuilder
from immuneML.workflows.instructions.MLProcess import MLProcess


class HPSelection:

    @staticmethod
    def update_split_count(state: TrainMLModelState, train_val_dataset):
        if state.selection.split_strategy == SplitType.LOOCV:
            state.selection.split_count = train_val_dataset.get_example_count()

        return state

    @staticmethod
    def run_selection(state: TrainMLModelState, train_val_dataset, current_path: Path, split_index: int) -> TrainMLModelState:

        path = HPSelection.create_selection_path(state, current_path)
        state = HPSelection.update_split_count(state, train_val_dataset)
        train_datasets, val_datasets = HPUtil.split_data(train_val_dataset, state.selection, path, state.label_configuration)

        n_labels = state.label_configuration.get_label_count()

        for idx, label in enumerate(state.label_configuration.get_label_objects()):

            print_log(f"Hyperparameter optimization: running the inner loop of nested CV: selection for label {label.name} "
                      f"(label {idx + 1} / {n_labels}).\n", include_datetime=True)

            selection_state = HPSelectionState(train_datasets, val_datasets, path, state.hp_strategy)
            state.assessment_states[split_index].label_states[label.name].selection_state = selection_state

            hp_setting = selection_state.hp_strategy.generate_next_setting()
            while hp_setting is not None:
                performance = HPSelection.evaluate_hp_setting(state, hp_setting, train_datasets, val_datasets,
                                                              path, label, split_index)
                hp_setting = selection_state.hp_strategy.generate_next_setting(hp_setting, performance)

            HPUtil.run_selection_reports(state, train_val_dataset, train_datasets, val_datasets, selection_state)

            print_log(f"Hyperparameter optimization: running the inner loop of nested CV: completed selection for "
                      f"label {label.name} (label {idx + 1} / {n_labels}).\n", include_datetime=True)

        return state

    @staticmethod
    def evaluate_hp_setting(state: TrainMLModelState, hp_setting: HPSetting, train_datasets: list, val_datasets: list,
                            current_path: Path, label: Label, assessment_split_index: int):

        performances = []
        for index in range(state.selection.split_count):
            performance = HPSelection.run_setting(state, hp_setting, train_datasets[index], val_datasets[index], index + 1,
                                                  current_path / f"split_{index + 1}" / f"{label.name}_{hp_setting.get_key()}",
                                                  label, assessment_split_index)
            performances.append(performance)

        return HPUtil.get_average_performance(performances)

    @staticmethod
    def run_setting(state: TrainMLModelState, hp_setting, train_dataset, val_dataset, split_index: int,
                    current_path: Path, label: Label, assessment_index: int):

        hp_item = MLProcess(train_dataset=train_dataset, test_dataset=val_dataset,
                            encoding_reports=state.selection.reports.encoding_reports.values(),
                            label_config=LabelConfiguration([label]), report_context=state.context,
                            number_of_processes=state.number_of_processes, metrics=state.metrics,
                            optimization_metric=state.optimization_metric, hp_setting=hp_setting,
                            ml_reports=state.selection.reports.model_reports.values(), label=label, path=current_path,
                            example_weighting=state.example_weighting).run(split_index)

        state.assessment_states[assessment_index].label_states[label.name].selection_state.hp_items[hp_setting.get_key()].append(hp_item)

        return hp_item.performance[state.optimization_metric.name.lower()] if hp_item.performance is not None else None

    @staticmethod
    def create_selection_path(state: TrainMLModelState, current_path: Path) -> Path:
        path = current_path / f"selection_{state.selection.split_strategy.name.lower()}"
        PathBuilder.build(path)
        return path



import plotly.graph_objects as go
from functools import lru_cache

import numpy as np
import pandas as pd


def KL(sequences, model_1, model_2):
    """
    Computes the KL divergence between two models (model_1 and model_2) for a given set of sequences.

    Args:
        sequences: list of sequences
        model_1: model 1
        model_2: model 2

    Returns:
        KL divergence value
    """
    return np.mean(get_kl_weights(model_1, model_2, sequences))


def get_kl_weights(model_1, model_2, sequences):
    return model_1.log_prob(sequences) - model_2.log_prob(sequences)


def evaluate_similarities(true_sequences, simulated_sequences, estimator):
    true_model = estimator(true_sequences)
    simulated_model = estimator(simulated_sequences)
    return KL(true_sequences, true_model, simulated_model), \
           KL(simulated_sequences, simulated_model, true_model)


class KLEvaluator:
    def __init__(self, true_sequences, simulated_sequences, estimator, n_sequences):
        self.true_sequences = true_sequences
        self.simulated_sequences = simulated_sequences
        self.true_model = estimator(true_sequences)
        self.simulated_model = estimator(simulated_sequences)
        self.estimator = estimator
        self._n_sequences = n_sequences

    @lru_cache()
    def true_kl_weights(self):
        return get_kl_weights(self.true_model, self.simulated_model, self.true_sequences)

    @lru_cache()
    def simulated_kl_weights(self):
        return get_kl_weights(self.simulated_model, self.true_model, self.simulated_sequences)

    def true_kl(self):
        return np.mean(self.true_kl_weights())

    def simulated_kl(self):
        return np.mean(self.simulated_kl_weights())

    def get_worst_true_sequences(self, n=20):
        indices = np.argsort(self.true_kl_weights())[-n:][::-1]
        return pd.DataFrame({"sequence": self.true_sequences[indices].tolist(),
                             "kl": self.true_kl_weights()[indices]})
        # return self.true_sequences[indices]

    def get_worst_simulated_sequences(self, n=20):
        indices = self._simulated_indices(n)
        return pd.DataFrame({"sequence": self.simulated_sequences[indices].tolist(),
                                "kl": self.simulated_kl_weights()[indices]})

    def _simulated_indices(self, n):
        return np.argsort(self.simulated_kl_weights())[-n:][::-1]

    def simulated_plot(self):
        n_sequences = self._n_sequences
        indices = self._simulated_indices(n_sequences)
        kmers = self.simulated_sequences[indices]
        weights = self.simulated_model.kmer_model.log_prob(kmers) - self.true_model.kmer_model.log_prob(kmers)
        scores = self.simulated_kl_weights()[indices]
        fig = self.get_plot(indices, kmers, scores, weights)
        fig.update_layout(yaxis_title="Sequence", xaxis_title="Position",
                          title="KL weights for the generated sequences that don't fit with the original model")
        return fig

    def original_plot(self):
        n_sequences = self._n_sequences
        indices = np.argsort(self.true_kl_weights())[-n_sequences:][::-1]
        kmers = self.true_sequences[indices]
        weights = self.true_model.kmer_model.log_prob(kmers) - self.simulated_model.kmer_model.log_prob(kmers)
        scores = self.true_kl_weights()[indices]
        fig = self.get_plot(indices, kmers, scores, weights)
        fig.update_layout(yaxis_title="Sequence", xaxis_title="Position",
                          title="KL weights for the original sequences that don't fit the generated model")
        # add label for colorbar
        #fig.update_layout(colorbar={"title": 'Your title'})

        return fig


    def get_plot(self, indices, kmers, scores, weights):
        n_sequences = len(indices)
        width = np.max(weights.shape[-1])
        z = np.full((n_sequences, width), np.nan)
        text = [['' for _ in range(width)] for _ in range(n_sequences)]
        sequences = []
        for kmers, weight, row, text_row, score in zip(kmers, weights, z, text, scores):
            row[:len(kmers)] = weight
            text_row[:len(kmers)] = [kmer.to_string() for kmer in kmers]

            seq = ''.join([kmers[0].to_string()[:-1]] + [kmer.to_string()[-1] for kmer in kmers])
            sequences.append(f'{seq} : {score:.2f}')
        fig = go.Figure(data=go.Heatmap(
            y=sequences,
            z=z,
            text=text,
            texttemplate="%{text}",
            colorbar=dict(title="KL weights"),
            colorscale=["green", 'yellow', 'red']
        ))
        fig.update_yaxes(autorange="reversed")
        return fig


from pathlib import Path

from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.ml_methods.util.Util import Util as MLUtil
from immuneML.presentation.TemplateParser import TemplateParser
from immuneML.presentation.html.Util import Util
from immuneML.util.PathBuilder import PathBuilder
from immuneML.util.StringHelper import StringHelper
from immuneML.workflows.instructions.exploratory_analysis.ExploratoryAnalysisState import ExploratoryAnalysisState


class ExploratoryAnalysisHTMLBuilder:
    """
    A class that will make a HTML file(s) out of ExploratoryAnalysisState object to show what analysis took place in
    the ExploratoryAnalysisInstruction.
    """

    CSS_PATH = EnvironmentSettings.html_templates_path / "css/custom.css"

    @staticmethod
    def build(state: ExploratoryAnalysisState) -> Path:
        """
        Function that builds the HTML files based on the ExploratoryAnalysis state.
        Arguments:
            state: ExploratoryAnalysisState object with details and results of the instruction
        Returns:
             path to the main HTML file (which is located under state.result_path)
        """
        base_path = PathBuilder.build(state.result_path / "../HTML_output/")
        html_map = ExploratoryAnalysisHTMLBuilder.make_html_map(state, base_path)
        result_file = base_path / f"ExploratoryAnalysis_{state.name}.html"

        TemplateParser.parse(template_path=EnvironmentSettings.html_templates_path / "ExploratoryAnalysis.html",
                             template_map=html_map, result_path=result_file)

        return result_file

    @staticmethod
    def make_html_map(state: ExploratoryAnalysisState, base_path: Path) -> dict:
        html_map = {
            "css_style": Util.get_css_content(ExploratoryAnalysisHTMLBuilder.CSS_PATH),
            "full_specs": Util.get_full_specs_path(base_path),
            "logfile": Util.get_logfile_path(base_path),
            'immuneML_version': MLUtil.get_immuneML_version(),
            "analyses": [
                {
                    **Util.make_dataset_html_map(analysis.dataset),
                    **{
                        "name": name,
                        "preprocessing_sequence": [
                            {
                                "preprocessing_name": preprocessing.__class__.__name__,
                                "preprocessing_params": ", ".join(
                                    [f"{key}: {value}" for key, value in vars(preprocessing).items()])
                            } for preprocessing in analysis.preprocessing_sequence
                        ] if analysis.preprocessing_sequence is not None else [],
                        "show_preprocessing": analysis.preprocessing_sequence is not None and len(
                            analysis.preprocessing_sequence) > 0,
                        "show_labels": analysis.label_config is not None and len(
                            analysis.label_config.get_labels_by_name()) > 0,
                        "analysis_labels": [{"name": label.name, "values": str(label.values)[1:-1]}
                                   for label in
                                   analysis.label_config.get_label_objects()] if analysis.label_config else None,
                        "encoding_key": analysis.encoder.name if analysis.encoder is not None else None,
                        "encoding_name": StringHelper.camel_case_to_word_string(
                            type(analysis.encoder).__name__) if analysis.encoder is not None
                        else None,
                        "encoding_params": [{"param_name": key, "param_value": str(value)} for key, value in
                                            vars(analysis.encoder).items()] if analysis.encoder is not None else None,
                        "show_encoding": analysis.encoder is not None,
                        "report": Util.to_dict_recursive(Util.update_report_paths(analysis.report_result, base_path),
                                                         base_path) if analysis.report_result is not None else None
                    }
                } for name, analysis in state.exploratory_analysis_units.items()]
        }

        for analysis in html_map["analyses"]:
            if analysis["report"] is not None:
                analysis["show_tables"] = len(analysis["report"]["output_tables"]) > 0 if "output_tables" in analysis["report"] else False
                analysis["show_text"] = len(analysis["report"]["output_text"]) > 0 if "output_text" in analysis["report"] else False
                analysis["show_info"] = analysis["report"]["info"] is not None and len(analysis["report"]["info"]) > 0 if "info" in analysis["report"] else False
            else:
                analysis["show_tables"] = False
                analysis["show_text"] = False
                analysis["show_info"] = False

        return html_map



import plotly.graph_objects as go


class PlotlyUtil:

    @staticmethod
    def add_single_axis_labels(figure, x_label, y_label, x_label_position, y_label_position):
        """
        Takes a multi-facet plotly figure and replaces the repetitive x and y axis labels
        with single axis labels in the form of annotations.

        Args:
            figure: a plotly figure
            x_label: the x label text
            y_label: the y label text
            x_label_position: the position of the new axis labels relative to the respective axes
            y_label_position: the position of the new axis labels relative to the respective axes

        Returns:
             an updated plotly figure

        """
        # hide subplot y-axis titles and x-axis titles
        for axis in figure.layout:
            if type(figure.layout[axis]) == go.layout.YAxis:
                figure.layout[axis].title.text = ''
            if type(figure.layout[axis]) == go.layout.XAxis:
                figure.layout[axis].title.text = ''

        # keep all other annotations and add single y-axis and x-axis title:
        figure.update_layout(
            # keep the original annotations and add a list of new annotations:
            annotations=list(figure.layout.annotations) +
                        [go.layout.Annotation(x=y_label_position, y=0.5, font={"size": 14},
                                              showarrow=False, text=y_label, textangle=-90, xref="paper",
                                              yref="paper")] +
                        [go.layout.Annotation(x=0.5, y=x_label_position, font={"size": 15},
                                              showarrow=False, text=x_label, textangle=-0,
                                              xref="paper", yref="paper")])
        return figure

    @staticmethod
    def get_amino_acid_color_map():
        '''To be used whenever plotting for example a barplot where each amino acid is represented,
        to be used as a value for plotly's color_discrete_map'''
        return {'Y': 'rgb(102, 197, 204)', 'W': 'rgb(179,222,105)', 'V': 'rgb(220, 176, 242)',
                'T': 'rgb(217,217,217)', 'S': 'rgb(141,211,199)', 'R': 'rgb(251,128,114)',
                'Q': 'rgb(158, 185, 243)', 'P': 'rgb(248, 156, 116)', 'N': 'rgb(135, 197, 95)',
                'M': 'rgb(254, 136, 177)', 'L': 'rgb(201, 219, 116)', 'K': 'rgb(255,237,111)',
                'I': 'rgb(180, 151, 231)', 'H': 'rgb(246, 207, 113)', 'G': 'rgb(190,186,218)',
                'F': 'rgb(128,177,211)', 'E': 'rgb(253,180,98)',  'D': 'rgb(252,205,229)',
                'C': 'rgb(188,128,189)', 'A': 'rgb(204,235,197)'}


import copy
import inspect
from pathlib import Path
from typing import List

import sklearn.metrics as sklearn_metrics

from immuneML.data_model.SequenceParams import RegionType
from immuneML.dsl.DefaultParamsLoader import DefaultParamsLoader
from immuneML.dsl.instruction_parsers.LabelHelper import LabelHelper
from immuneML.dsl.symbol_table.SymbolTable import SymbolTable
from immuneML.dsl.symbol_table.SymbolType import SymbolType
from immuneML.environment.LabelConfiguration import LabelConfiguration
from immuneML.environment.SequenceType import SequenceType
from immuneML.hyperparameter_optimization.config.LeaveOneOutConfig import LeaveOneOutConfig
from immuneML.hyperparameter_optimization.config.ManualSplitConfig import ManualSplitConfig
from immuneML.hyperparameter_optimization.config.SplitConfig import SplitConfig
from immuneML.hyperparameter_optimization.config.SplitType import SplitType
from immuneML.ml_methods.clustering.ClusteringMethod import ClusteringMethod
from immuneML.ml_methods.dim_reduction.DimRedMethod import DimRedMethod
from immuneML.reports.Report import Report
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.workflows.instructions.clustering.ClusteringInstruction import ClusteringInstruction
from immuneML.workflows.instructions.clustering.clustering_run_model import ClusteringSetting


class ClusteringParser:

    def parse(self, key: str, instruction: dict, symbol_table: SymbolTable, path: Path = None) -> ClusteringInstruction:
        valid_keys = [k for k in inspect.signature(ClusteringInstruction.__init__).parameters.keys()
                      if k not in ['result_path', 'name', 'self', 'label_config']] + ['type', 'labels']

        ParameterValidator.assert_keys(instruction.keys(), valid_keys, ClusteringParser.__name__, key)

        ParameterValidator.assert_region_type(instruction, ClusteringParser.__name__)
        ParameterValidator.assert_sequence_type(instruction, ClusteringParser.__name__)
        if instruction['validation_type'] is not None:
            ParameterValidator.assert_type_and_value(instruction['validation_type'], list, ClusteringParser.__name__,
                                                     'validation_type')
            for val_type in instruction['validation_type']:
                ParameterValidator.assert_in_valid_list(val_type, ['method_based', 'result_based'],
                                                        ClusteringParser.__name__, 'validation_type')

        dataset = symbol_table.get(instruction['dataset'])
        clustering_settings = parse_clustering_settings(key, instruction, symbol_table)
        metrics = parse_metrics(key, instruction, symbol_table)
        label_config = parse_labels(key, instruction, dataset)
        reports = parse_reports(key, instruction, symbol_table)
        split_config = parse_split_config(instruction)

        return ClusteringInstruction(dataset=dataset, metrics=metrics, clustering_settings=clustering_settings,
                                     name=key, label_config=label_config, reports=reports,
                                     sequence_type=SequenceType[instruction['sequence_type'].upper()],
                                     region_type=RegionType[instruction['region_type'].upper()],
                                     split_config=split_config, validation_type=instruction['validation_type'])


def parse_split_config(instruction) -> SplitConfig:
    try:
        split_key = 'split_config'
        default_params = DefaultParamsLoader.load("instructions/", SplitConfig.__name__)
        instruction[split_key] = {**default_params, **instruction[split_key]}

        valid_keys = [k for k in inspect.signature(SplitConfig.__init__).parameters.keys() if k not in ['self']]

        ParameterValidator.assert_keys(instruction[split_key].keys(), valid_keys, ClusteringParser.__name__, split_key,
                                       exclusive=False)
        ParameterValidator.assert_type_and_value(instruction[split_key]['split_count'], int, ClusteringParser.__name__,
                                                 'split_count', min_inclusive=1)

        split_strategy = SplitType[instruction[split_key]["split_strategy"].upper()]
        training_percentage = float(
            instruction[split_key]["training_percentage"]) if split_strategy == SplitType.RANDOM else -1

        if split_strategy == SplitType.RANDOM:
            assert 0. <= training_percentage <= 1., \
                f'{ClusteringParser.__name__}: training_percentage has to between 0 and 1 if split_strategy is RANDOM.'

        elif split_strategy == SplitType.MANUAL:
            ParameterValidator.assert_keys(keys=instruction[split_key]["manual_config"].keys(),
                                           valid_keys=["discovery_data", "validation_data"],
                                           location=ClusteringParser.__name__, parameter_name="manual_config",
                                           exclusive=True)

            ParameterValidator.assert_valid_tabular_file(instruction[split_key]["manual_config"]["discovery_data"],
                                                         location=ClusteringParser.__name__,
                                                         parameter_name="discovery_data")

            ParameterValidator.assert_valid_tabular_file(instruction[split_key]["manual_config"]["validation_data"],
                                                         location=ClusteringParser.__name__,
                                                         parameter_name="validation_data")

        return SplitConfig(split_strategy=split_strategy,
                           split_count=instruction[split_key]['split_count'], training_percentage=training_percentage,
                           manual_config=ManualSplitConfig(train_metadata_path=instruction[split_key]['discovery_data'],
                                                           test_metadata_path=instruction[split_key]['validation_data'])
                           if "manual_config" in instruction[split_key] else None,
                           leave_one_out_config=LeaveOneOutConfig(**instruction[split_key]["leave_one_out_config"])
                           if "leave_one_out_config" in instruction[split_key] else None)

    except KeyError as key_error:
        raise KeyError(
            f"{ClusteringParser.__name__}: parameter {key_error.args[0]} was not defined under {split_key}.") from key_error


def parse_labels(key, instruction, dataset) -> LabelConfiguration:
    if 'labels' in instruction and instruction['labels'] is not None:
        label_config = LabelHelper.create_label_config(instruction['labels'], dataset, key, 'labels')
    else:
        label_config = LabelConfiguration()
    return label_config


def parse_reports(key, instruction, symbol_table) -> List[Report]:
    if 'reports' not in instruction or instruction['reports'] is None:
        return []
    else:
        ParameterValidator.assert_type_and_value(instruction['reports'], list, 'ClusteringParser', 'reports')
        valid_reports = symbol_table.get_keys_by_type(SymbolType.REPORT)

        ParameterValidator.assert_all_type_and_value(instruction['reports'], str, 'ClusteringParser', 'reports')
        ParameterValidator.assert_all_in_valid_list(instruction['reports'], valid_reports, 'ClusteringParser',
                                                    'reports')

        reports = [symbol_table.get(report_id) for report_id in instruction['reports']]
        return reports


def parse_metrics(key: str, instruction: dict, symbol_table: SymbolTable) -> List[str]:
    ParameterValidator.assert_type_and_value(instruction['metrics'], list, 'ClusteringParser', f'{key}:metrics')
    ParameterValidator.assert_all_type_and_value(instruction['metrics'], str, 'ClusteringParser', f'{key}:metrics')

    for metric in instruction['metrics']:
        assert hasattr(sklearn_metrics, metric), (f"Clustering parser: metric {metric} is not a valid metric. See the "
                                                  f"list of scikit-learn's metrics for clustering.")

    return instruction['metrics']


def parse_clustering_settings(key: str, instruction: dict, symbol_table: SymbolTable) -> List[ClusteringSetting]:
    ParameterValidator.assert_type_and_value(instruction['clustering_settings'], list, 'ClusteringParser',
                                             'key:clustering_settings')
    valid_encodings = symbol_table.get_keys_by_type(SymbolType.ENCODING)
    valid_dim_red = [method.symbol for method in symbol_table.get_by_type(SymbolType.ML_METHOD)
                     if isinstance(method.item, DimRedMethod)]
    valid_clusterings = [method.symbol for method in symbol_table.get_by_type(SymbolType.ML_METHOD)
                         if isinstance(method.item, ClusteringMethod)]

    settings_objs = []
    for setting in instruction['clustering_settings']:
        setting_obj = make_setting_obj(setting, valid_encodings, valid_clusterings, valid_dim_red, symbol_table,
                                       instruction)
        settings_objs.append(setting_obj)

    return settings_objs


def make_setting_obj(setting, valid_encodings, valid_clusterings, valid_dim_red, symbol_table, instruction):
    ParameterValidator.assert_keys_present(setting.keys(), ['encoding', 'method'], 'ClusteringParser',
                                           'clustering_settings')
    ParameterValidator.assert_keys(setting.keys(), ['encoding', 'dim_reduction', 'method'], 'ClusteringParser',
                                   'clustering_settings', exclusive=False)

    ParameterValidator.assert_in_valid_list(setting['encoding'], valid_encodings, 'ClusteringParser', 'encoding')
    ParameterValidator.assert_in_valid_list(setting['method'], valid_clusterings, 'ClusteringParser',
                                            'method')
    if 'dim_reduction' in setting and setting['dim_reduction'] is not None:
        ParameterValidator.assert_in_valid_list(setting['dim_reduction'], valid_dim_red, 'ClusteringParser',
                                                'dim_reduction')
        dim_reduction = copy.deepcopy(symbol_table.get(setting['dim_reduction']))
        dim_red_params = symbol_table.get_config(setting['dim_reduction'])
        dim_red_name = setting['dim_reduction']
    else:
        dim_reduction, dim_red_params, dim_red_name = None, None, None

    encoder = make_encoder_obj(symbol_table, setting['encoding'], instruction['dataset'])
    method = copy.deepcopy(symbol_table.get(setting['method']))

    return ClusteringSetting(encoder=encoder, encoder_params=symbol_table.get_config(setting['encoding']),
                             encoder_name=setting['encoding'], clustering_method=method,
                             clustering_params=symbol_table.get_config(setting['method']),
                             clustering_method_name=setting['method'], dim_reduction_method=dim_reduction,
                             dim_red_params=dim_red_params, dim_red_name=dim_red_name)


def make_encoder_obj(symbol_table, encoding_key, dataset_key):
    return symbol_table.get(encoding_key).build_object(symbol_table.get(dataset_key),
                                                       **symbol_table.get_config(encoding_key)["encoder_params"]) \
        .set_context({"dataset": symbol_table.get(dataset_key)})



import logging
from pathlib import Path

import pandas as pd
import plotly.express as px

from immuneML.data_model import bnp_util
from immuneML.data_model.AIRRSequenceSet import AIRRSequenceSet
from immuneML.data_model.SequenceParams import RegionType
from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.data_model.datasets.ElementDataset import ReceptorDataset, SequenceDataset
from immuneML.data_model.datasets.RepertoireDataset import RepertoireDataset
from immuneML.data_model.SequenceSet import ReceptorSequence
from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.environment.SequenceType import SequenceType
from immuneML.reports.PlotlyUtil import PlotlyUtil
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.ReportUtil import ReportUtil
from immuneML.reports.data_reports.DataReport import DataReport
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder
from immuneML.util.PositionHelper import PositionHelper


class AminoAcidFrequencyDistribution(DataReport):
    """
    Generates a barplot showing the relative frequency of each amino acid at each position in the sequences of a dataset.

    Example output:

    .. image:: _static/images/reports/amino_acid_frequency.png
       :alt: Amino acid frequency
       :width: 800

    .. image:: ../../_static/images/reports/amino_acid_frequency_change.png
       :alt: Amino acid frequency change
       :width: 800

    **Specification arguments:**

    - alignment (str): Alignment style for aligning sequences of different lengths. Options are as follows:

      - CENTER: center-align sequences of different lengths. The middle amino acid of any sequence be labelled position 0. By default, alignment is CENTER.

      - LEFT: left-align sequences of different lengths, starting at 0.

      - RIGHT: right align sequences of different lengths, ending at 0 (counting towards negative numbers).

      - IMGT: align sequences based on their IMGT positional numbering, considering the sequence region_type (IMGT_CDR3 or IMGT_JUNCTION).
        The main difference between CENTER and IMGT is that IMGT aligns the first and last amino acids, adding gaps in the middle,
        whereas CENTER aligns the middle of the sequences, padding with gaps at the start and end of the sequence.
        When region_type is IMGT_JUNCTION, the IMGT positions run from 104 (conserved C) to 118 (conserved W/F). When IMGT_CDR3 is used, these positions are 105 to 117.
        For long CDR3 sequences, additional numbers are added in between IMGT positions 111 and 112.
        See the official IMGT documentation for more details: https://www.imgt.org/IMGTScientificChart/Numbering/CDR3-IMGTgaps.html

    - relative_frequency (bool): Whether to plot relative frequencies (true) or absolute counts (false) of the
      positional amino acids. Note that when sequences are of different length, setting relative_frequency to True will
      produce different results depending on the alignment type, as some positions are only covered by the longest sequences.
      By default, relative_frequency is False.

    - split_by_label (bool): Whether to split the plots by a label. If set to true, the Dataset must either contain a
      single label, or alternatively the label of interest can be specified under 'label'. If split_by_label is set to
      true, the percentage-wise frequency difference between classes is plotted additionally. By default,
      split_by_label is False.

    - label (str): if split_by_label is set to True, a label can be specified here.

    - region_type (str): which part of the sequence to check; e.g., IMGT_CDR3

    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            reports:
                my_aa_freq_report:
                    AminoAcidFrequencyDistribution:
                        relative_frequency: False
                        split_by_label: True
                        label: CMV
                        region_type: IMGT_CDR3

    """

    @classmethod
    def build_object(cls, **kwargs):
        location = AminoAcidFrequencyDistribution.__name__

        if "imgt_positions" in kwargs:
            raise ValueError(
                f"{location}: parameter 'imgt_positions' is deprecated. For 'imgt_positions: True', use 'alignment: IMGT'. For 'imgt_positions: False', use any other alignment option (CENTER/LEFT/RIGHT).")

        ParameterValidator.assert_type_and_value(kwargs["alignment"], str, location, "alignment")
        ParameterValidator.assert_in_valid_list(kwargs["alignment"].upper(), ["IMGT", "LEFT", "RIGHT", "CENTER"],
                                                location, "alignment")
        ParameterValidator.assert_type_and_value(kwargs["relative_frequency"], bool, location, "relative_frequency")
        ParameterValidator.assert_type_and_value(kwargs["split_by_label"], bool, location, "split_by_label")
        ParameterValidator.assert_region_type(kwargs, location)

        ReportUtil.update_split_by_label_kwargs(kwargs, location)

        return AminoAcidFrequencyDistribution(**{**kwargs, 'region_type': RegionType[kwargs['region_type'].upper()]})

    def __init__(self, dataset: Dataset = None, alignment: bool = None, relative_frequency: bool = None,
                 split_by_label: bool = None, label: str = None, region_type: RegionType = RegionType.IMGT_CDR3,
                 result_path: Path = None, number_of_processes: int = 1, name: str = None):
        super().__init__(dataset=dataset, result_path=result_path, number_of_processes=number_of_processes, name=name)
        self.alignment = alignment
        self.relative_frequency = relative_frequency
        self.split_by_label = split_by_label
        self.label_name = label
        self.region_type = region_type

    def _generate(self) -> ReportResult:
        PathBuilder.build(self.result_path)
        self.label_name = self._get_label_name()

        freq_dist = self._get_plotting_data()

        tables = []
        figures = []

        tables.append(self._write_output_table(freq_dist,
                                               self.result_path / "amino_acid_frequency_distribution.tsv",
                                               name="Table of amino acid frequencies"))

        figures.append(self._safe_plot(freq_dist=freq_dist, plot_callable="_plot_distribution"))

        if self.split_by_label:
            frequency_change = self._compute_frequency_change(freq_dist)

            tables.append(self._write_output_table(frequency_change,
                                                   self.result_path / f"frequency_change.tsv",
                                                   name=f"Frequency change between classes"))
            figures.append(self._safe_plot(frequency_change=frequency_change, plot_callable="_plot_frequency_change"))

        return ReportResult(name=self.name,
                            info="A barplot showing the relative frequency of each amino acid at each position in "
                                 "the sequences of a dataset.",
                            output_figures=[fig for fig in figures if fig is not None],
                            output_tables=[table for table in tables if table is not None])

    def _get_plotting_data(self):
        if isinstance(self.dataset, SequenceDataset):
            plotting_data = self._get_sequence_dataset_plotting_data()

        elif isinstance(self.dataset, ReceptorDataset):
            plotting_data = self._get_receptor_dataset_plotting_data()

        elif isinstance(self.dataset, RepertoireDataset):
            plotting_data = self._get_repertoire_dataset_plotting_data()

        if not self.split_by_label:
            plotting_data.drop(columns=[self.label_name], inplace=True)

        return plotting_data

    def _get_sequence_dataset_plotting_data(self):
        return self._count_dict_per_class_to_df(self._count_aa_frequencies(self.dataset.data, class_name=None))

    def _count_dict_per_class_to_df(self, raw_count_dict_per_class):
        result_dfs = []

        for class_name, raw_count_dict in raw_count_dict_per_class.items():
            result_df = self._count_dict_to_df(raw_count_dict)
            result_df[self.label_name] = class_name
            result_dfs.append(result_df)

        return pd.concat(result_dfs)

    def _get_receptor_dataset_plotting_data(self):
        result_dfs = []

        data = self.dataset.data
        chains = list(set(data.locus.tolist()))

        for chain in chains:
            mask = [el == chain for el in data.locus.tolist()]
            raw_count_dict_per_class = self._count_aa_frequencies(data[mask])

            for class_name, raw_count_dict in raw_count_dict_per_class.items():
                result_df = self._count_dict_to_df(raw_count_dict)
                result_df["locus"] = chain
                result_df[self.label_name] = class_name
                result_dfs.append(result_df)

        return pd.concat(result_dfs)

    def _get_repertoire_dataset_plotting_data(self):
        raw_count_dict_per_class = {}
        label_name = self._get_label_name()

        if self.split_by_label:
            class_names = self.dataset.get_metadata([label_name])[label_name]
        else:
            class_names = [0] * self.dataset.get_example_count()

        for repertoire, class_name in zip(self.dataset.get_data(), class_names):
            self._count_aa_frequencies(repertoire.data, raw_count_dict_per_class, class_name)

        return self._count_dict_per_class_to_df(raw_count_dict_per_class)

    def _count_aa_frequencies(self, data: AIRRSequenceSet, raw_count_dict=None, class_name: str = None):
        raw_count_dict = {} if raw_count_dict is None else raw_count_dict

        for item in data.to_iter():
            sequence = getattr(item, bnp_util.get_sequence_field_name(self.region_type, SequenceType.AMINO_ACID))
            seq_pos = self._get_positions(len(sequence))

            if self.split_by_label and self.label_name is not None:
                if class_name is None:
                    cls_name = getattr(item, self.label_name)
                else:
                    cls_name = class_name
            else:
                cls_name = 0

            if cls_name not in raw_count_dict:
                raw_count_dict[cls_name] = {}

            for aa, pos in zip(sequence, seq_pos):
                if pos not in raw_count_dict[cls_name]:
                    raw_count_dict[cls_name][pos] = {legal_aa: 0 for legal_aa in
                                                     EnvironmentSettings.get_sequence_alphabet(
                                                         SequenceType.AMINO_ACID)}

                raw_count_dict[cls_name][pos][aa] += 1

        return raw_count_dict

    def _count_dict_to_df(self, raw_count_dict):
        df_dict = {"amino acid": [], "position": [], "count": [], "relative frequency": []}
        for pos in raw_count_dict:
            df_dict["position"].extend([pos] * 20)
            df_dict["amino acid"].extend(list(raw_count_dict[pos].keys()))

            counts = list(raw_count_dict[pos].values())
            total_count_for_pos = sum(counts)

            df_dict["count"].extend(counts)
            df_dict["relative frequency"].extend([count / total_count_for_pos for count in counts])

        return pd.DataFrame(df_dict)

    def _get_positions(self, sequence_length: int):
        if self.alignment == 'IMGT':
            positions = PositionHelper.gen_imgt_positions_from_length(sequence_length, self.region_type)
        elif self.alignment == 'LEFT':
            positions = list(range(1, sequence_length + 1))
        elif self.alignment == "RIGHT":
            positions = list(range(-sequence_length + 2, 1))
        else:  # self.alignment == "CENTER
            positions = list(range(1, sequence_length + 1))
            positions = [pos - round(max(positions) / 2) for pos in positions]

        return [str(pos) for pos in positions]

    def _write_results_table(self, results_table):
        file_path = self.result_path / "amino_acid_frequency_distribution.tsv"

        results_table.to_csv(file_path, index=False)

        return ReportOutput(path=file_path, name="Table of amino acid frequencies")

    def _plot_distribution(self, freq_dist):
        freq_dist.sort_values(by=["amino acid"], ascending=False, inplace=True)
        category_orders = None if self.label_name not in freq_dist.columns \
            else {self.label_name: sorted(set(freq_dist[self.label_name]))}

        y = "relative frequency" if self.relative_frequency else "count"

        figure = px.bar(freq_dist, x="position", y=y, color="amino acid", text="amino acid",
                        facet_col=self.label_name if self.label_name in freq_dist.columns else None,
                        facet_row="locus" if "locus" in freq_dist.columns else None,
                        color_discrete_map=PlotlyUtil.get_amino_acid_color_map(),
                        category_orders=category_orders,
                        labels={"position": "IMGT position" if self.alignment == "IMGT" else "Position",
                                "count": "Count",
                                "relative frequency": "Relative frequency",
                                "amino acid": "Amino acid"}, template="plotly_white")
        figure.update_xaxes(categoryorder='array', categoryarray=self._get_position_order(freq_dist["position"]))
        figure.update_layout(showlegend=False, yaxis={'categoryorder': 'category ascending'})

        if self.relative_frequency:
            figure.update_yaxes(tickformat=",.0%", range=[0, 1])

        file_path = self.result_path / "amino_acid_frequency_distribution.html"
        figure.write_html(str(file_path))

        return ReportOutput(path=file_path, name="Amino acid frequency distribution")

    def _get_position_order(self, positions):
        if self.alignment == "IMGT":
            if min(positions) == "105" and max(positions) == "117":
                return PositionHelper.gen_imgt_positions_from_cdr3_length(len(set(positions)))
            elif min(positions) == "104" and max(positions) == "118":
                return ["104"] + PositionHelper.gen_imgt_positions_from_cdr3_length(len(set(positions)) - 2) + ["118"]
        else:
            return [str(pos) for pos in sorted(set(positions.astype(int)))]

    def _compute_frequency_change(self, freq_dist):
        classes = sorted(set(freq_dist[self.label_name]))
        assert len(classes) == 2, \
            (f"{AminoAcidFrequencyDistribution.__name__}: cannot compute frequency change when the number of "
             f"classes is not 2: {classes}")

        class_a_df = freq_dist[freq_dist[self.label_name] == classes[0]]
        class_b_df = freq_dist[freq_dist[self.label_name] == classes[1]]

        on = ["amino acid", "position"]
        on = on + ["locus"] if "locus" in freq_dist.columns else on

        merged_dfs = pd.merge(class_a_df, class_b_df, on=on, how="outer", suffixes=["_a", "_b"])
        merged_dfs = merged_dfs[(merged_dfs["relative frequency_a"] + merged_dfs["relative frequency_b"]) > 0]

        merged_dfs["frequency_change"] = merged_dfs["relative frequency_a"] - merged_dfs["relative frequency_b"]

        pos_class_a = merged_dfs[merged_dfs["frequency_change"] > 0]
        pos_class_b = merged_dfs[merged_dfs["frequency_change"] < 0]

        pos_class_a["positive_class"] = classes[0]
        pos_class_b["positive_class"] = classes[1]
        pos_class_b["frequency_change"] = 0 - pos_class_b["frequency_change"]

        keep_cols = on + ["frequency_change", "positive_class"]
        pos_class_a = pos_class_a[keep_cols]
        pos_class_b = pos_class_b[keep_cols]

        return pd.concat([pos_class_a, pos_class_b])

    def _plot_frequency_change(self, frequency_change):
        figure = px.bar(frequency_change, x="position", y="frequency_change", color="amino acid", text="amino acid",
                        facet_col="positive_class",
                        facet_row="locus" if "locus" in frequency_change.columns else None,
                        color_discrete_map=PlotlyUtil.get_amino_acid_color_map(),
                        labels={"position": "IMGT position" if self.alignment == "IMGT" else "Position",
                                "positive_class": self.label_name,
                                "frequency_change": "Difference in relative frequency",
                                "amino acid": "Amino acid"}, template="plotly_white")

        figure.update_xaxes(categoryorder='array', categoryarray=self._get_position_order(frequency_change["position"]))
        figure.update_layout(showlegend=False, yaxis={'categoryorder': 'category ascending'})

        figure.update_yaxes(tickformat=",.0%")

        file_path = self.result_path / "frequency_change.html"
        figure.write_html(str(file_path))

        return ReportOutput(path=file_path, name="Frequency difference between amino acid usage in the two classes")

    def _get_label_name(self):
        if self.split_by_label:
            if self.label_name is None:
                return list(self.dataset.get_label_names())[0]
            else:
                return self.label_name
        else:
            return None

    def check_prerequisites(self):
        if self.split_by_label:
            if self.label_name is None:
                if len(self.dataset.get_label_names()) != 1:
                    logging.warning(f"{AminoAcidFrequencyDistribution.__name__}: ambiguous label: split_by_label was "
                                    f"set to True but no label name was specified, and the number of available labels "
                                    f"is {len(self.dataset.get_label_names())}: {self.dataset.get_label_names()}. "
                                    f"Skipping this report...")
                    return False
            elif self.label_name not in self.dataset.get_label_names():
                logging.warning(f"{AminoAcidFrequencyDistribution.__name__}: the specified label name "
                                f"({self.label_name}) was not available among the dataset labels: "
                                f"{self.dataset.get_label_names()}. Skipping this report...")
                return False

        return True



import copy
import os
from multiprocessing.pool import Pool
from pathlib import Path
from uuid import uuid4

import pandas as pd

from immuneML.data_model import bnp_util
from immuneML.data_model.AIRRSequenceSet import AIRRSequenceSet
from immuneML.data_model.SequenceParams import RegionType
from immuneML.data_model.bnp_util import write_yaml, read_yaml
from immuneML.data_model.datasets.ElementDataset import SequenceDataset
from immuneML.data_model.datasets.RepertoireDataset import RepertoireDataset
from immuneML.data_model.SequenceSet import Repertoire
from immuneML.environment.SequenceType import SequenceType
from immuneML.preprocessing.filters.CountAggregationFunction import CountAggregationFunction
from immuneML.preprocessing.filters.Filter import Filter
from immuneML.util.ParameterValidator import ParameterValidator
from scripts.specification_util import update_docs_per_mapping


class DuplicateSequenceFilter(Filter):
    """
    Collapses duplicate nucleotide or amino acid sequences within each repertoire in the given RepertoireDataset or within a SequenceDataset.
    This filter can be applied to Repertoires, RepertoireDatasets, and SequenceDatasets.

    Sequences are considered duplicates if the following fields are identical:

    - amino acid or nucleotide sequence (whichever is specified)

    - v and j genes (note that the full field including subgroup + gene is used for matching, i.e. V1 and V1-1 are not considered duplicates)

    - chain

    - region type

    For all other fields (the non-specified sequence type, custom lists, sequence identifier) only the first occurring
    value is kept.

    Note that this means the count value of a sequence with a given sequence identifier might not be the same as before
    removing duplicates, unless count_agg = FIRST is used.

    **Specification arguments:**

    - filter_sequence_type (:py:obj:`~immuneML.environment.SequenceType.SequenceType`): Whether the sequences should be collapsed on the nucleotide or amino acid level. Valid options are defined by the SequenceType enum.

    - region_type (str): which part of the sequence to examine, by default, this is IMGT_CDR3

    - count_agg (:py:obj:`~immuneML.preprocessing.filters.CountAggregationFunction.CountAggregationFunction`): determines how the sequence counts of duplicate sequences are aggregated. Valid options are defined by the CountAggregationFunction enum.


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        preprocessing_sequences:
            my_preprocessing:
                - my_filter:
                    DuplicateSequenceFilter:
                        # required parameters:
                        filter_sequence_type: AMINO_ACID
                        # optional parameters (if not specified the values bellow will be used):
                        batch_size: 4
                        count_agg: SUM
                        region_type: IMGT_CDR3

    """

    @classmethod
    def build_object(cls, **kwargs):
        location = cls.__name__
        ParameterValidator.assert_keys(kwargs.keys(), ["filter_sequence_type", "batch_size", "count_agg", "region_type"], location,
                                       "DuplicateSequenceFilter")
        ParameterValidator.assert_in_valid_list(kwargs["filter_sequence_type"].upper(), [item.name for item in SequenceType],
                                                location, "filter_sequence_type")
        ParameterValidator.assert_in_valid_list(kwargs["region_type"].upper(), [item.name for item in RegionType],
                                                location, "region_type")
        ParameterValidator.assert_in_valid_list(kwargs["count_agg"].upper(), [item.name for item in CountAggregationFunction], location,
                                                "count_agg")
        ParameterValidator.assert_type_and_value(kwargs["batch_size"], int, location, "batch_size", 1)
        return DuplicateSequenceFilter(filter_sequence_type=SequenceType[kwargs["filter_sequence_type"].upper()],
                                       region_type=RegionType[kwargs['region_type'].upper()],
                                       batch_size=kwargs["batch_size"], count_agg=CountAggregationFunction[kwargs["count_agg"].upper()])

    def __init__(self, filter_sequence_type: SequenceType, batch_size: int, count_agg: CountAggregationFunction,
                 result_path: Path = None, region_type: RegionType = RegionType.IMGT_CDR3):
        super().__init__(result_path)
        self.filter_sequence_type = filter_sequence_type
        self.region_type = region_type
        self.count_agg = count_agg
        self.batch_size = batch_size

        self.sequence_of_interest = bnp_util.get_sequence_field_name(self.region_type, self.filter_sequence_type)
        self.sequence_to_ignore = bnp_util.get_sequence_field_name(self.region_type, [t for t in SequenceType if t != self.filter_sequence_type][0])

    def process_dataset(self, dataset, result_path: Path, number_of_processes=1):
        self.result_path = result_path if result_path is not None else self.result_path

        self.check_dataset_type(dataset, [RepertoireDataset, SequenceDataset], "DuplicateSequenceFilter")

        processed_dataset = dataset.clone()

        if isinstance(dataset, RepertoireDataset):
            with Pool(self.batch_size) as pool:
                repertoires = pool.map(self._process_repertoire, dataset.repertoires)

            processed_dataset.repertoires = repertoires
            return processed_dataset

        elif isinstance(dataset, SequenceDataset):
            processed_dataset = self._process_sequences(dataset)
            return processed_dataset

    def _prepare_group_by_field(self, columns):
        rep_fields = list(AIRRSequenceSet.get_field_type_dict().keys())
        groupby_fields = copy.deepcopy(rep_fields)
        if self.sequence_to_ignore in groupby_fields:
            groupby_fields.remove(self.sequence_to_ignore)
        groupby_fields.remove("duplicate_count")
        groupby_fields.remove("sequence_id")
        groupby_fields.remove("cell_id")

        for field in set(rep_fields).difference(set(columns)):
            if field in groupby_fields:
                groupby_fields.remove(field)

        return groupby_fields

    def _prepare_agg_dict(self, columns, custom_lists):

        agg_dict = {"sequence_id": "first"}

        if self.sequence_to_ignore in columns:
            agg_dict[self.sequence_to_ignore] = "first"

        if "duplicate_count" in columns:
            agg_dict["duplicate_count"] = self.count_agg.value

        if "cell_id" in columns:
            agg_dict["cell_id"] = "first"

        for key in custom_lists:
            agg_dict[key] = "first"

        return agg_dict

    def _process_repertoire(self, repertoire: Repertoire) -> Repertoire:
        no_duplicates = self._filter_duplicates(repertoire)

        no_duplicates.to_csv(f"{self.result_path}/{repertoire.data_filename.stem}_filtered.tsv", sep='\t', index=False)
        write_yaml(Path(f"{self.result_path}/{repertoire.metadata_filename.stem}_filtered.yaml"), repertoire.metadata)

        return Repertoire(data_filename=Path(f"{self.result_path}/{repertoire.data_filename.stem}_filtered.tsv"),
                          metadata_filename=Path(f"{self.result_path}/{repertoire.metadata_filename.stem}_filtered.yaml"),
                          identifier=str(uuid4().hex),
                          dynamic_fields=repertoire.dynamic_fields)

    def _process_sequences(self, dataset: SequenceDataset) -> SequenceDataset:
        no_duplicates = self._filter_duplicates(dataset)
        metadata = read_yaml(dataset.dataset_file)

        os.makedirs(self.result_path, exist_ok=True)
        no_duplicates.to_csv(f"{self.result_path}/{dataset.filename.stem}_filtered.tsv", sep='\t', index=False)
        write_yaml(Path(f"{self.result_path}/{dataset.dataset_file.stem}_filtered.yaml"), metadata)

        return SequenceDataset(name=dataset.name,
                               filename=Path(f"{self.result_path}/{dataset.filename.stem}_filtered.tsv"),
                               dataset_file=Path(f"{self.result_path}/{dataset.dataset_file.stem}_filtered.yaml"),
                               dynamic_fields=dataset.dynamic_fields)

    def _filter_duplicates(self, dataset):
        data = dataset.data.topandas()
        data['duplicate_count'].replace(-1, pd.NA, inplace=True)
        columns = data.columns

        groupby_fields = self._prepare_group_by_field(columns)
        custom_lists = list(dataset.dynamic_fields.keys())
        agg_dict = self._prepare_agg_dict(columns, custom_lists)

        no_duplicates = data.groupby(groupby_fields, sort=False).agg(agg_dict).reset_index()
        return no_duplicates

    @staticmethod
    def get_documentation():
        doc = str(DuplicateSequenceFilter.__doc__)

        mapping = {
            "Valid options are defined by the CountAggregationFunction enum.": f"Valid values are: {[e.name.lower() for e in CountAggregationFunction]}.",
            "Valid options are defined by the SequenceType enum.": f"Valid values are: {[e.name.lower() for e in SequenceType]}."
        }

        doc = update_docs_per_mapping(doc, mapping)

        return doc



import abc
import logging
from pathlib import Path

from immuneML.analysis.data_manipulation.DataReshaper import DataReshaper
from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.encoding_reports.EncodingReport import EncodingReport
from immuneML.util.PathBuilder import PathBuilder


class FeatureReport(EncodingReport):
    """
    Base class for reports that plot something about the reshaped feature values of any dataset.
    """

    def __init__(self, dataset: Dataset = None, result_path: Path = None,
                 color_grouping_label: str = None, row_grouping_label=None, column_grouping_label=None,
                 name: str = None, number_of_processes: int = 1):
        super().__init__(dataset=dataset, result_path=result_path, name=name, number_of_processes=number_of_processes)
        self.x = "feature"
        self.color = color_grouping_label
        self.facet_row = row_grouping_label
        self.facet_column = column_grouping_label

    def _generate_report_result(self) -> ReportResult:
        PathBuilder.build(self.result_path)
        data_long_format = DataReshaper.reshape(self.dataset, self.dataset.get_label_names())
        table_result = self._write_results_table(data_long_format)
        report_output = self._safe_plot(data_long_format=data_long_format)
        output_tables = [table_result]
        if report_output is None:
            output_figures = None
        elif isinstance(report_output, tuple):
            output_figures = report_output[0] if isinstance(report_output[0], list) else [report_output[0]]
            output_tables = report_output[1]
        else:
            output_figures = report_output if isinstance(report_output, list) else [report_output]
        return ReportResult(name=self.name, output_figures=output_figures, output_tables=output_tables)

    def _write_results_table(self, data) -> ReportOutput:
        table_path = self.result_path / f"feature_values.csv"
        data.to_csv(table_path, index=False)
        return ReportOutput(table_path, "feature values")

    def std(self, x):
        return x.std(ddof=0)

    @abc.abstractmethod
    def _plot(self, data_long_format) -> ReportOutput:
        pass

    def check_prerequisites(self):
        location = self.__class__.__name__
        run_report = True

        if self.dataset.encoded_data is None or self.dataset.encoded_data.examples is None:
            logging.warning(
                f"{location}: this report can only be created for an encoded dataset. {location} report will not be created.")
            run_report = False
        elif len(self.dataset.encoded_data.examples.shape) != 2:
            logging.warning(
                f"{location}: this report can only be created for a 2-dimensional encoded dataset. {location} report will not be created.")
            run_report = False
        else:
            legal_labels = list(self.dataset.get_label_names())

            labels = [self.color, self.facet_row, self.facet_column]

            for label_param in labels:
                if label_param is not None:
                    if label_param not in legal_labels:
                        logging.warning(
                            f"{location}: undefined label '{label_param}'. Legal options are: {legal_labels}. {location} report will not be created.")
                        run_report = False

        return run_report



import argparse
import glob
import itertools as it
import logging
import sys
import warnings
from pathlib import Path

import yaml

from immuneML.api.galaxy.Util import Util
from immuneML.encodings.kmer_frequency.sequence_encoding.SequenceEncodingType import SequenceEncodingType
from immuneML.ml_methods.classifiers.MLMethod import MLMethod
from immuneML.reports.ml_reports.CoefficientPlottingSetting import CoefficientPlottingSetting
from immuneML.util.PathBuilder import PathBuilder
from immuneML.util.ReadsType import ReadsType
from immuneML.util.ReflectionHandler import ReflectionHandler


def get_sequence_enc_type(sequence_type, position_type, gap_type):
    if sequence_type == "complete":
        encoding_type = SequenceEncodingType.IDENTITY
    else:
        if position_type == "positional":
            if gap_type == "gapped":
                encoding_type = SequenceEncodingType.IMGT_GAPPED_KMER
            else:
                encoding_type = SequenceEncodingType.IMGT_CONTINUOUS_KMER
        else:
            if gap_type == "gapped":
                encoding_type = SequenceEncodingType.GAPPED_KMER
            else:
                encoding_type = SequenceEncodingType.CONTINUOUS_KMER

    return encoding_type.name


def build_encodings_specs(args):
    encodings = dict()

    for i in range(len(args.sequence_type)):
        enc_name = f"encoding_{i + 1}"
        enc_spec = dict()

        enc_spec["sequence_encoding"] = get_sequence_enc_type(args.sequence_type[i],
                                                              None if args.position_type is None else args.position_type[i],
                                                              None if args.gap_type is None else args.gap_type[i])
        enc_spec["reads"] = args.reads[i]

        if args.sequence_type[i] == "subsequence":
            if args.gap_type[i] == "gapped":
                enc_spec["k_left"] = args.k_left[i]
                enc_spec["k_right"] = args.k_right[i]
                enc_spec["min_gap"] = args.min_gap[i]
                enc_spec["max_gap"] = args.max_gap[i]
            else:
                enc_spec["k"] = args.k[i]

        encodings[enc_name] = {"KmerFrequency": enc_spec}

    return encodings


def get_ml_method_spec(ml_method_class, model_selection_n_folds=5):
    if ml_method_class == "LogisticRegression" or ml_method_class == "SimpleLogisticRegression":
        ml_spec = {
            "logistic_regression": {
                "LogisticRegression": {
                    "penalty": ["l1"],
                    "C": [0.01, 0.1, 1, 10, 100],
                    "class_weight": ["balanced"],
                    "show_warnings": False
                },
                "model_selection_cv": True,
                "model_selection_n_folds": model_selection_n_folds
            }
        }
    elif ml_method_class == "RandomForestClassifier":
        ml_spec = {
            "random_forest": {
                "RandomForestClassifier": {
                    "n_estimators": [10, 50, 100],
                    "class_weight": ["balanced"],
                    "show_warnings": False
                },
                "model_selection_cv": True,
                "model_selection_n_folds": model_selection_n_folds
            }
        }
    elif ml_method_class == "SVM":
        ml_spec = {
            "support_vector_machine": {
                "SVC": {
                    "penalty": ["l1"],
                    "dual": False,
                    "C": [0.01, 0.1, 1, 10, 100],
                    "class_weight": ["balanced"],
                    "show_warnings": False
                },
                "model_selection_cv": True,
                "model_selection_n_folds": model_selection_n_folds
            }
        }
    elif ml_method_class == "KNN":
        ml_spec = {
            "k_nearest_neighbors": {
                "KNN": {
                    "n_neighbors": [3, 5, 7],
                    "show_warnings": False
                },
                "model_selection_cv": True,
                "model_selection_n_folds": model_selection_n_folds
            }
        }
    else:
        ml_spec = {ml_method_class: ml_method_class}

    return ml_spec


def build_ml_methods_specs(args):
    ml_methods_spec = dict()

    for method in args.ml_methods:
        ml_methods_spec.update(get_ml_method_spec(method))

    return ml_methods_spec


def build_settings_specs(enc_names, ml_names):
    return [{"encoding": enc_name, "ml_method": ml_name} for enc_name, ml_name in it.product(enc_names, ml_names)]


def build_labels(labels_str):
    labels = labels_str.split(",")
    return [label.strip().strip("'\"") for label in labels]


def build_specs(args):
    enc_specs = build_encodings_specs(args)
    ml_specs = build_ml_methods_specs(args)
    settings_specs = build_settings_specs(enc_specs.keys(), ml_specs.keys())
    labels = build_labels(args.labels)

    specs = {
        "definitions": {
            "datasets": {
                "dataset": {
                    "format": "AIRR",
                    "params": {"dataset_file": Util.discover_dataset_path()}
                }
            },
            "encodings": enc_specs,
            "ml_methods": ml_specs,
            "reports": {
                "coefficients": {
                    "Coefficients": {
                        "coefs_to_plot": [CoefficientPlottingSetting.N_LARGEST.name],
                        "n_largest": [25]
                    }
                },
                "benchmark": "MLSettingsPerformance"
            }
        },
        "instructions": {
            "inst1": {
                "type": "TrainMLModel",
                "settings": settings_specs,
                "assessment": {
                    "split_strategy": "random",
                    "split_count": args.split_count,
                    "training_percentage": args.training_percentage / 100,
                    "reports": {
                        "models": ["coefficients"]
                    }
                },
                "selection": {
                    "split_strategy": "random",
                    "split_count": 1,
                    "training_percentage": 0.7,
                },
                "labels": labels,
                "dataset": "dataset",
                "strategy": "GridSearch",
                "metrics": ["accuracy", "balanced_accuracy", "precision", "recall", "auc"],
                "number_of_processes": 8,
                "reports": ["benchmark"],
                "optimization_metric": "accuracy",
                'refit_optimal_model': True
            }
        }
    }


    return specs


def check_arguments(args):
    assert 100 >= args.training_percentage >= 10, "training_percentage must range between 10 and 100"
    assert args.split_count >= 1, "The minimal split_count is 1."

    encoding_err = "When multiple encodings are used, fields must still be of equal length, add 'NA' variables where necessary"
    assert len(args.sequence_type) == len(args.reads), encoding_err
    assert args.position_type is None or len(args.sequence_type) == len(args.position_type), encoding_err
    assert args.gap_type is None or len(args.sequence_type) == len(args.gap_type), encoding_err
    assert args.k is None or len(args.sequence_type) == len(args.k), encoding_err
    assert args.k_left is None or len(args.sequence_type) == len(args.k_left), encoding_err
    assert args.k_right is None or len(args.sequence_type) == len(args.k_right), encoding_err
    assert args.min_gap is None or len(args.sequence_type) == len(args.min_gap), encoding_err
    assert args.max_gap is None or len(args.sequence_type) == len(args.max_gap), encoding_err


def parse_commandline_arguments(args):
    ReflectionHandler.get_classes_by_partial_name("", "ml_methods/classifiers/")
    ml_method_names = [cl.__name__ for cl in ReflectionHandler.all_nonabstract_subclasses(MLMethod)] + ["SimpleLogisticRegression"]

    parser = argparse.ArgumentParser(description="tool for building immuneML Galaxy YAML from arguments")
    parser.add_argument("-o", "--output_path", required=True, help="Output location for the generated yaml file (directiory).")
    parser.add_argument("-f", "--file_name", default="specs.yaml",
                        help="Output file name for the yaml file. Default name is 'specs.yaml' if not specified.")
    parser.add_argument("-l", "--labels", required=True,
                        help="Which metadata labels should be predicted for the dataset (separated by comma).")
    parser.add_argument("-m", "--ml_methods", nargs="+", choices=ml_method_names, required=True,
                        help="Which machine learning methods should be applied.")
    parser.add_argument("-t", "--training_percentage", type=float, required=True,
                        help="The percentage of data used for training.")
    parser.add_argument("-c", "--split_count", type=int, required=True,
                        help="The number of times to repeat the training process with a different random split of the data.")
    parser.add_argument("-s", "--sequence_type", choices=["complete", "subsequence"], default=["subsequence"], nargs="+",
                        help="Whether complete CDR3 sequences are used, or k-mer subsequences.")
    parser.add_argument("-p", "--position_type", choices=["invariant", "positional"], nargs="+",
                        help="Whether IMGT-positional information is used for k-mers, or the k-mer positions are position-invariant.")
    parser.add_argument("-g", "--gap_type", choices=["gapped", "ungapped"], nargs="+", help="Whether the k-mers contain gaps.")
    parser.add_argument("-k", "--k", type=int, nargs="+", help="K-mer size.")
    parser.add_argument("-kl", "--k_left", type=int, nargs="+", help="Length before gap when k-mers are used.")
    parser.add_argument("-kr", "--k_right", type=int, nargs="+", help="Length after gap when k-mers are used.")
    parser.add_argument("-gi", "--min_gap", type=int, nargs="+", help="Minimal gap length when gapped k-mers are used.")
    parser.add_argument("-ga", "--max_gap", type=int, nargs="+", help="Maximal gap length when gapped k-mers are used.")
    parser.add_argument("-r", "--reads", choices=[ReadsType.UNIQUE.value, ReadsType.ALL.value], nargs="+", default=[ReadsType.UNIQUE.value],
                        help="Whether k-mer counts should be scaled by unique clonotypes or all observed receptor sequences")

    return parser.parse_args(args)


def main(args):
    logging.basicConfig(filename="build_yaml_from_args_log.txt", level=logging.INFO, format='%(asctime)s %(levelname)s: %(message)s')
    warnings.showwarning = lambda message, category, filename, lineno, file=None, line=None: logging.warning(message)

    parsed_args = parse_commandline_arguments(args)
    check_arguments(parsed_args)

    specs = build_specs(parsed_args)

    PathBuilder.build(parsed_args.output_path)
    output_location = Path(parsed_args.output_path) / parsed_args.file_name

    with output_location.open("w") as file:
        yaml.dump(specs, file)

    return output_location


if __name__ == "__main__":
    main(sys.argv[1:])



import pandas as pd

from immuneML.IO.dataset_import.DataImport import DataImport
from immuneML.data_model.SequenceParams import RegionType
from scripts.specification_util import update_docs_per_mapping


class IGoRImport(DataImport):
    """
    Imports data generated by `IGoR <https://github.com/qmarcou/IGoR>`_ simulations into a Repertoire-, or SequenceDataset.
    RepertoireDatasets should be used when making predictions per repertoire, such as predicting a disease state.
    SequenceDatasets should be used when predicting values for unpaired (single-chain) immune receptors, like
    antigen specificity.

    Note that you should run IGoR with the --CDR3 option specified, this tool imports the generated CDR3 files.
    Sequences with missing anchors are not imported, meaning only sequences with value '1' in the anchors_found column are imported.
    Nucleotide sequences are automatically translated to amino acid sequences.

    Reference: Quentin Marcou, Thierry Mora, Aleksandra M. Walczak
    ‘High-throughput immune repertoire analysis with IGoR’. Nature Communications, (2018)
    `doi.org/10.1038/s41467-018-02832-w <https://doi.org/10.1038/s41467-018-02832-w>`_.

    **Specification arguments:**

    - path (str): For RepertoireDatasets, this is the path to a directory with IGoR files to import. For Sequence- or
      ReceptorDatasets this path may either be the path to the file to import, or the path to the folder locating one
      or multiple files with .tsv, .csv or .txt extensions. By default path is set to the current working directory.

    - is_repertoire (bool): If True, this imports a RepertoireDataset. If False, it imports a SequenceDataset.
      By default, is_repertoire is set to True.

    - metadata_file (str): Required for RepertoireDatasets. This parameter specifies the path to the metadata file.
      This is a csv file with columns filename, subject_id and arbitrary other columns which can be used as labels in
      instructions. Only the IGoR files included under the column 'filename' are imported into the RepertoireDataset.
      For setting Sequence- or ReceptorDataset labels, metadata_file is ignored, use label_columns instead.

    - label_columns (list): For Sequence- or ReceptorDataset, this parameter can be used to explicitly set the column
      names of labels to import. These labels can be used as prediction target. When label_columns are not set, label
      names are attempted to be discovered automatically (any column name which is not used in the column_mapping).
      For setting RepertoireDataset labels, label_columns is ignored, use metadata_file instead.

    - import_with_stop_codon (bool): Whether sequences with stop codons should be included in the imported sequences.
      By default, import_with_stop_codon is False.

    - import_out_of_frame (bool): Whether out of frame sequences (with value '0' in column is_inframe) should be
      included in the imported sequences. By default, import_out_of_frame is False.

    - import_illegal_characters (bool): Whether to import sequences that contain illegal characters, i.e., characters
      that do not appear in the sequence alphabet (amino acids including stop codon '*', or nucleotides). When set to
      false, filtering is only applied to the sequence type of interest (when running immuneML in amino acid mode, only
      entries with illegal characters in the amino acid sequence are removed). By default, import_illegal_characters
      is False.

    - import_empty_nt_sequences (bool): imports sequences which have an empty nucleotide sequence field; can be True
      or False. By default, import_empty_nt_sequences is set to True.

    - region_type (str): Which part of the sequence to check when importing. By default, this value is set to IMGT_CDR3.
      This means the first and last amino acids are removed from the CDR3 sequence, as IGoR uses the IMGT junction.
      Specifying any other value will result in importing the sequences as they are. Valid values for region_type are
      the names of the :py:obj:`~immuneML.data_model.receptor.RegionType.RegionType` enum.

    - column_mapping (dict): A mapping from IGoR column names to immuneML's internal data representation. A custom column mapping can be specified here if necessary (for example; adding additional data fields if they are present in the IGoR file, or using alternative column names). Valid immuneML fields that can be specified here are defined by the AIRR standard (AIRRSequenceSet). For IGoR, this is by default set to:

        .. indent with spaces
        .. code-block:: yaml

            nt_CDR3: cdr3
            seq_index: sequence_id

    - separator (str): Column separator, for IGoR this is by default ",".


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            datasets:
                my_igor_dataset:
                    format: IGoR
                    params:
                        path: path/to/files/
                        is_repertoire: True # whether to import a RepertoireDataset (True) or a SequenceDataset (False)
                        metadata_file: path/to/metadata.csv # metadata file for RepertoireDataset
                        import_with_stop_codon: False # whether to include sequences with stop codon in the dataset
                        import_out_of_frame: False # whether to include out of frame sequences in the dataset
                        import_illegal_characters: False # remove sequences with illegal characters for the sequence_type being used
                        import_empty_nt_sequences: True # keep sequences even though the nucleotide sequence might be empty
                        # Optional fields with IGoR-specific defaults, only change when different behavior is required:
                        separator: "," # column separator
                        region_type: IMGT_CDR3 # what part of the sequence to import
                        column_mapping: # column mapping IGoR: immuneML
                            nt_CDR3: cdr3
                            seq_index: sequence_id
                            igor_column_name1: metadata_label1
                            igor_column_name2: metadata_label2

    """
    CODON_TABLE = {
        'ATA': 'I', 'ATC': 'I', 'ATT': 'I', 'ATG': 'M',
        'ACA': 'T', 'ACC': 'T', 'ACG': 'T', 'ACT': 'T',
        'AAC': 'N', 'AAT': 'N', 'AAA': 'K', 'AAG': 'K',
        'AGC': 'S', 'AGT': 'S', 'AGA': 'R', 'AGG': 'R',
        'CTA': 'L', 'CTC': 'L', 'CTG': 'L', 'CTT': 'L',
        'CCA': 'P', 'CCC': 'P', 'CCG': 'P', 'CCT': 'P',
        'CAC': 'H', 'CAT': 'H', 'CAA': 'Q', 'CAG': 'Q',
        'CGA': 'R', 'CGC': 'R', 'CGG': 'R', 'CGT': 'R',
        'GTA': 'V', 'GTC': 'V', 'GTG': 'V', 'GTT': 'V',
        'GCA': 'A', 'GCC': 'A', 'GCG': 'A', 'GCT': 'A',
        'GAC': 'D', 'GAT': 'D', 'GAA': 'E', 'GAG': 'E',
        'GGA': 'G', 'GGC': 'G', 'GGG': 'G', 'GGT': 'G',
        'TCA': 'S', 'TCC': 'S', 'TCG': 'S', 'TCT': 'S',
        'TTC': 'F', 'TTT': 'F', 'TTA': 'L', 'TTG': 'L',
        'TAC': 'Y', 'TAT': 'Y', 'TAA': '*', 'TAG': '*',
        'TGC': 'C', 'TGT': 'C', 'TGA': '*', 'TGG': 'W',
    }

    def preprocess_file(self, df: pd.DataFrame) -> pd.DataFrame:

        if 'junction' in df.columns and ('junction_aa' not in df.columns or all(df.junction_aa == "")):
            df['junction_aa'] = df['junction'].apply(IGoRImport.translate_sequence)

        if "cdr3" in df.columns and "cdr3_aa" not in df.columns:
            df["cdr3_aa"] = df["cdr3"].apply(IGoRImport.translate_sequence)

        if 'vj_in_frame' in df.columns:
            df['vj_in_frame'] = df['vj_in_frame'].astype(str).replace('1', 'T').replace('0', 'F')

        if 'anchors_found' in df.columns:
            df = df[df.anchors_found == 1]

        return df

    @staticmethod
    def translate_sequence(nt_seq):
        """
        Code inspired by: https://github.com/prestevez/dna2proteins/blob/master/dna2proteins.py
        """
        aa_seq = []
        end = len(nt_seq) - (len(nt_seq) % 3) - 1
        for i in range(0, end, 3):
            codon = nt_seq[i:i + 3]
            if codon in IGoRImport.CODON_TABLE:
                aminoacid = IGoRImport.CODON_TABLE[codon]
                aa_seq.append(aminoacid)
            else:
                aa_seq.append("_")
        return "".join(aa_seq)

    @staticmethod
    def get_documentation():
        doc = str(IGoRImport.__doc__)

        region_type_values = str([region_type.name for region_type in RegionType])[1:-1].replace("'", "`")

        mapping = {
            "Valid values for region_type are the names of the :py:obj:`~immuneML.data_model.receptor.RegionType.RegionType` enum.": f"Valid values are {region_type_values}.",
            "Valid immuneML fields that can be specified here are defined by the AIRR standard (AIRRSequenceSet)": f"Valid immuneML fields that can be specified here by `the AIRR Rearrangement Schema <https://docs.airr-community.org/en/latest/datarep/rearrangements.html>`_."
        }
        doc = update_docs_per_mapping(doc, mapping)
        return doc



import logging

import pandas as pd

from immuneML.caching.CacheHandler import CacheHandler
from immuneML.data_model.datasets.ElementDataset import ReceptorDataset
from immuneML.data_model.SequenceParams import RegionType


class TCRdistHelper:

    @staticmethod
    def compute_tcr_dist(dataset: ReceptorDataset, label_names: list, cores: int = 1):
        return CacheHandler.memo_by_params((('dataset_identifier', dataset.identifier), ("type", "TCRrep")),
                                           lambda: TCRdistHelper._compute_tcr_dist(dataset, label_names, cores))

    @staticmethod
    def _compute_tcr_dist(dataset: ReceptorDataset, label_names: list, cores: int):
        """
        Computes the tcrdist distances by creating a TCRrep object and calling compute_distances() function.

        Parameters `ntrim` and `ctrim` in TCRrep object for CDR3 are adjusted to account for working with IMGT CDR3 definition if IMGT CDR3 was set
        as region_type for the dataset upon importing. `deduplicate` parameter is set to False as we assume that we work with clones in immuneML,
        and not individual receptors.

        Args:
            dataset: receptor dataset for which all pairwise distances between receptors will be computed
            label_names: a list of label names (e.g., specific epitopes) to be used for later classification or reports
            cores: how many cpus to use for computation

        Returns:
            an instance of TCRrep object with computed pairwise distances between all receptors in the dataset

        """
        from tcrdist.repertoire import TCRrep

        df = TCRdistHelper.prepare_tcr_dist_dataframe(dataset, label_names)
        tcr_rep = TCRrep(cell_df=df, chains=['alpha', 'beta'], organism=dataset.labels["organism"], cpus=cores,
                         deduplicate=False,
                         compute_distances=False)

        if 'region_type' not in dataset.labels:
            logging.warning(
                f"{TCRdistHelper.__name__}: Parameter 'region_type' was not set for dataset {dataset.name}, keeping default tcrdist "
                f"values for parameters 'ntrim' and 'ctrim'. For more information, see tcrdist3 documentation. To avoid this warning, "
                f"set the region type when importing the dataset.")
        elif dataset.labels['region_type'] == RegionType.IMGT_CDR3:
            tcr_rep.kargs_a['cdr3_a_aa']['ntrim'] = 2
            tcr_rep.kargs_a['cdr3_a_aa']['ctrim'] = 1
            tcr_rep.kargs_b['cdr3_b_aa']['ntrim'] = 2
            tcr_rep.kargs_b['cdr3_b_aa']['ctrim'] = 1
        elif dataset.labels['region_type'] != RegionType.IMGT_JUNCTION:
            raise RuntimeError(
                f"{TCRdistHelper.__name__}: TCRdist metric can be computed only if IMGT_CDR3 or IMGT_JUNCTION are used as region "
                f"types, but for dataset {dataset.name}, it is set to {dataset.labels['region_type']} instead.")

        tcr_rep.compute_distances()

        return tcr_rep

    @staticmethod
    def add_default_allele_to_v_gene(v_gene: str):
        if v_gene is not None and "*" not in v_gene:
            return f"{v_gene}*01"
        else:
            return v_gene

    @staticmethod
    def prepare_tcr_dist_dataframe(dataset: ReceptorDataset, label_names: list) -> pd.DataFrame:
        if len(label_names) > 1:
            raise NotImplementedError(
                f"TCRdist: multiple labels specified ({str(label_names)[1:-1]}), but only single label binary class "
                f"is currently supported in immuneML.")

        df = dataset.data.topandas()

        epitope_name = 'epitope' if 'epitope' in df.columns else 'Epitope' if 'Epitope' in df.columns else ''

        if "subject" not in df:
            df['subject'] = "sub" + df['cell_id']

        df.loc[df['v_call'].str.contains("\*"), 'v_call'] = [TCRdistHelper.add_default_allele_to_v_gene(el) for el in
                                                             df.loc[df['v_call'].str.contains("\*"), 'v_call']]
        df.loc[df['j_call'].str.contains("\*"), 'j_call'] = [TCRdistHelper.add_default_allele_to_v_gene(el) for el in
                                                             df.loc[df['j_call'].str.contains("\*"), 'j_call']]

        df['clone_id'] = df['cell_id']
        cols_to_keep = ['cdr3', 'cdr3_aa', 'v_call', 'j_call', 'duplicate_count', 'subject', epitope_name, 'clone_id']

        df_alpha = (df[df.locus == 'TRA'][cols_to_keep]
                    .rename(columns={"cdr3_aa": "cdr3_a_aa", "cdr3": "cdr3_a_nucseq", "v_call": "v_a_gene",
                                     'j_call': "j_a_gene", "duplicate_count": "count"}))
        df_alpha.loc[:, 'count'] = [1 if el in [-1, None] else el for el in df_alpha['count']]

        df_beta = (df[df.locus == 'TRB'][cols_to_keep].rename(
            columns={"cdr3_aa": "cdr3_b_aa", "cdr3": "cdr3_b_nucseq", "v_call": "v_b_gene",
                     'j_call': "j_b_gene", "duplicate_count": "count"}))
        df_beta.loc[:, 'count'] = [1 if el in [-1, None] else el for el in df_beta['count']]

        df = df_alpha.merge(df_beta, on=['clone_id', epitope_name, 'subject', 'count'])

        return df



import abc
import pickle
import shutil
from pathlib import Path
from typing import List

import dill

from immuneML.IO.dataset_export.AIRRExporter import AIRRExporter
from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.encodings.EncoderParams import EncoderParams


class DatasetEncoder(metaclass=abc.ABCMeta):
    """

    **YAML specification:**

        encodings:
            e1: <encoder_class> # encoding without parameters

            e2:
                <encoder_class>: # encoding with parameters
                    parameter: value
    """
    def __init__(self, name: str = None):
        self.name = name

    @staticmethod
    @abc.abstractmethod
    def build_object(dataset: Dataset, **params):
        """
        Creates an instance of the relevant subclass of the DatasetEncoder class using the given parameters.
        This method will be called during parsing time (early in the immuneML run), such that parameters and dataset type can be tested here.

        The build_object method should do the following:

          1. Check parameters: immuneML should crash if wrong user parameters are specified. The ParameterValidator utility class may be used for parameter testing.

          2. Check the dataset type: immuneML should crash if the wrong dataset type is specified for this encoder. For example, DeepRCEncoder should only work for RepertoireDatasets and crash if the dataset is of another type.

          3. Create an instance of the correct Encoder class, using the given parameters. Return this object.
             Some encoders have different subclasses depending on the dataset type. Make sure to return an instance of the correct subclass.
             For instance: KmerFrequencyEncoder has different subclasses for each dataset type. When the dataset is a Repertoire dataset, KmerFreqRepertoireEncoder should be returned.


        Args:

            **params: keyword arguments that will be provided by users in the specification (if immuneML is used as a command line tool) or in the
             dictionary when calling the method from the code, and which should be used to create the Encoder object

        Returns:

            the object of the appropriate Encoder class

        """
        pass

    @abc.abstractmethod
    def encode(self, dataset, params: EncoderParams) -> Dataset:
        """
        This is the main encoding method of the Encoder. It takes in a given dataset, computes an EncodedData object,
        and returns a copy of the dataset with the attached EncodedData object.


        Args:

            dataset: A dataset object (Sequence, Receptor or RepertoireDataset)

            params: An EncoderParams object containing few utility parameters which may be used during encoding (e.g., number of parallel processes to use).

        Returns:

            A copy of the original dataset, with an EncodedData object added to the dataset.encoded_data field.

        """

        pass

    @staticmethod
    def load_encoder(encoder_file: Path):
        """
        The load_encoder method can load the encoder given the folder where the same class of the model was previously stored using the store function.
        Encoders are stored in pickle format. If the encoder uses additional files, they should be explicitly loaded here as well.

        If there are no additional files, this method does not need to be overwritten.
        If there are additional files, its contents should be as follows:

            encoder = DatasetEncoder.load_encoder(encoder_file)
            encoder.my_additional_file = DatasetEncoder.load_attribute(encoder, encoder_file, "my_additional_file")


        Arguments:

            encoder_file (Path): path to the encoder file where the encoder was stored using store() function

        Returns:

            the loaded Encoder object
        """
        with encoder_file.open("rb") as file:
            encoder = dill.load(file)
        return encoder

    @staticmethod
    def load_attribute(encoder, encoder_file: Path, attribute: str):
        """
        Utility method for loading correct file paths when loading an encoder (see: load_encoder).
        This method should not be overwritten.
        """
        if encoder_file is not None:
            file_path = encoder_file.parent / getattr(encoder, attribute).name
            setattr(encoder, attribute, file_path)
            assert getattr(encoder, attribute).is_file(), f"{type(encoder).__name__}: could not load {attribute} from {getattr(encoder, attribute)}."
        return encoder

    @staticmethod
    def store_encoder(encoder, encoder_file: Path):
        """
        The store_encoder function stores the given encoder such that it can be imported later using load function.
        It uses pickle to store the Python object, as well as the additional filenames which should be returned by
        the get_additional_files() method.

        This method should not be overwritten.


        Arguments:

            encoder: the encoder object

            encoder_file (Path): path to the encoder file

        Returns:

            the encoder file

        """
        with encoder_file.open("wb") as file:
            dill.dump(encoder, file)

        encoder_dir = encoder_file.parent
        for file in encoder.get_additional_files():
            shutil.copy(file, encoder_dir / file.name)

        return encoder_file

    @staticmethod
    def get_additional_files() -> List[str]:
        """
        Should return a list with all the files that need to be stored when storing the encoder.
        For example, SimilarToPositiveSequenceEncoder stores all 'positive' sequences in the training data,
        and predicts a sequence to be 'positive' if it is similar to any positive sequences in the training data.
        In that case, these positive sequences are stored in a file.

        For many encoders, it may not be necessary to store additional files.
        """
        return []

    def set_context(self, context: dict):
        """
        This method can be used to attach the full dataset (as part of a dictionary), as opposed to the dataset which
        is passed to the .encode() method. When training ML models, that data split is usually a training/validation
        subset of the total dataset.

        In most cases, an encoder should only use the 'dataset' argument passed to the .encode() method to compute
        the encoded data. Using information from the full dataset, which includes the test data, may result in data leakage.
        For example, some encoders normalise the computed feature values (e.g., KmerFrequencyEncoder). Such normalised
        feature values should be based only on the current data split, and test data should remain unseen.

        To avoid confusion about which version of the dataset to use, the full dataset is by default not attached,
        and attaching the full dataset should be done explicitly when required. For instance, if the encoded data is
        some kind of distance matrix (e.g., DistanceEncoder), the distance between examples in the training and test
        dataset should be included. Note that this does *not* entail data leakage: the test examples are not used to
        improve the computation of distances. The distances to test examples are determined by an algorithm which does
        not 'learn' from test data.

        To explicitly enable using the full dataset in the encoder, the contents of this method should be as follows:

        self.context = context
        return self


        Args:
            context: a dictionary containing the full dataset
        """
        return self

    def store(self, encoded_dataset, params: EncoderParams):
        """
        Stores the given encoded dataset. This method should not be overwritten.
        """
        AIRRExporter.export(encoded_dataset, params.result_path)



import sys
import argparse
import random
import numpy as np

# Ensure the immuneML/ project 'root dir' is added to sys.path
# Adding "." and "../" allows the script to be run from immuneML/ and immuneML/scripts/
# When encountering ModuleNotFoundError, try adding the absolute path to the project 'root dir' here
sys.path.extend([".", "../"])

from scripts.checker_util import *
from immuneML.data_model.encoded_data.EncodedData import EncodedData
from immuneML.dsl.DefaultParamsLoader import DefaultParamsLoader
from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.environment.Label import Label
from immuneML.ml_methods.classifiers.MLMethod import MLMethod
from immuneML.util.ReflectionHandler import ReflectionHandler


def parse_commandline_arguments(args):
    parser = argparse.ArgumentParser(description="Tool for testing new immuneML MLMethod classes")

    usage_args = parser.add_argument_group('usage arguments')
    usage_args.add_argument("-m", "--ml_method_file", type=str, required=True, help="Path to the MLMethod file, placed in the correct immuneML subfolder. ")
    usage_args.add_argument("-p", "--no_default_parameters",  action='store_true', help="If enabled, it is assumed that no default parameters file exists, and the MLMethod can be run without supplying additional parameters. ")

    logging_args = parser.add_argument_group('logging arguments')
    logging_args.add_argument("-l", "--log_file", type=str, default="check_new_ml_method_log.txt", help="Path to the output log file. If already present, the file will be overwritten (default='./check_new_ml_method_log.txt').")
    logging_args.add_argument("-t", "--tmp_path", type=str, default="./tmp", help="Path to the temporary output folder. If already present, the folder will be overwritten (default='./tmp').")

    return parser.parse_args(args)


def check_ml_method_file(ml_method_file):
    logging.info(f"Checking file name and location...")

    assert ml_method_file.parent.stem == "classifiers", f"Error: The new MLMethod file must be located inside a new package (folder) under 'immuneML/ml_methods/classifiers'. Found the following folder where 'classifiers' was expected: {ml_method_file.parent}"
    assert ml_method_file.parent.parent.stem == "ml_methods", f"Error: The new MLMethod file must be located inside a new package (folder) under 'immuneML/ml_methods/classifiers'. Found the following folder where 'ml_methods' was expected: {ml_method_file.parent}"

    assert ml_method_file.suffix == ".py", f"Error: '{ml_method_file}' is not a Python file"
    assert ml_method_file.stem[0].isupper(), f"Error: Filename '{ml_method_file}' does not start with an upper case character. Please rename the ML method (both class and filename) to '{ml_method_file.stem.title()}'"

    check_is_alphanum_name(ml_method_file.stem)

    logging.info(f"...Checking file name and location done.")

def get_ml_method_class(ml_method_file_path):
    logging.info(f"Attempting to load MLMethod class named '{ml_method_file_path.stem}'...")
    ml_method_class = ReflectionHandler.get_class_from_path(ml_method_file_path, class_name=ml_method_file_path.stem)
    logging.info(f"...Loading MLMethod class done.")

    return ml_method_class

def get_default_params_filepath_from_name(ml_method_file_path):
    return EnvironmentSettings.default_params_path / "ml_methods" / f"{DefaultParamsLoader.convert_to_snake_case(ml_method_file_path.stem)}_params.yaml"

def check_methods(ml_method_instance):
    logging.info("Checking methods...")
    mssg = "Error: class method '{}' should not be overwritten from MLMethod. Please remove this method."

    assert MLMethod.fit == ml_method_instance.__class__.fit, mssg.format("fit", ml_method_instance.__class__.fit)
    assert MLMethod._initialize_fit == ml_method_instance.__class__._initialize_fit, mssg.format("_initialize_fit")
    assert MLMethod.fit_by_cross_validation == ml_method_instance.__class__.fit_by_cross_validation, mssg.format("fit_by_cross_validation")
    assert MLMethod._assert_matching_label == ml_method_instance.__class__._assert_matching_label, mssg.format("_assert_matching_label")
    assert MLMethod.predict == ml_method_instance.__class__.predict, mssg.format("predict")
    assert MLMethod.predict_proba == ml_method_instance.__class__.predict_proba, mssg.format("predict_proba")
    assert MLMethod.get_feature_names == ml_method_instance.__class__.get_feature_names, mssg.format("get_feature_names")
    assert MLMethod.get_label_name == ml_method_instance.__class__.get_label_name, mssg.format("get_label_name")
    assert MLMethod.get_classes == ml_method_instance.__class__.get_classes, mssg.format("get_classes")
    assert MLMethod.get_positive_class == ml_method_instance.__class__.get_positive_class, mssg.format("get_positive_class")

    if MLMethod.check_encoder_compatibility != ml_method_instance.__class__.check_encoder_compatibility:
        logging.warning(f"class method 'check_encoder_compatibility' was overwritten from MLMethod. Please ensure this was intentional (for example: if more than just the Encoder type needs to be checked). ")

        # , mssg.format("check_encoder_compatibility", ml_method_instance.__class__.check_encoder_compatibility)


    check_base_vs_instance_methods(MLMethod, ml_method_instance)

    compatible_encoders = ml_method_instance.get_compatible_encoders()
    assert isinstance(compatible_encoders, list), f"get_compatible_encoders() should return a list of compatible encoder classes (found: {type(compatible_encoders)}). If no compatible encoder exists, it should return an empty list. "
    if ml_method_instance.get_compatible_encoders() == []:
        logging.warning(f"get_compatible_encoders() returned an empty list. At least one encoder class must be returned in this list in order to be able to use the MLMethod in practice.")

    assert isinstance(ml_method_instance.can_predict_proba(), bool), f"can_predict_proba() should return boolean value (found: {type(ml_method_instance.can_predict_proba())}). "
    assert isinstance(ml_method_instance.can_fit_with_example_weights(), bool), f"can_fit_with_example_weights() should return boolean value (found: {type(ml_method_instance.can_fit_with_example_weights())}). "

    logging.info("...Checking methods done.")

def get_example_encoded_data():
    """
    Note: by default, this random EncodedData object is used to represent an encoded dataset.
    Some MLMethods require a more specific encoding type and will fail with this example
    for instance: if different dimensions or additional info are used in the encoded data.

    You may overwrite this method to supply a custom EncodedData object which meets the requirements of your MLMethod.
    """
    enc_data = EncodedData(examples=np.array([[1, 0, 0], [0, 1, 1], [1, 1, 1], [0, 1, 1], [1, 0, 0], [0, 1, 1], [1, 1, 1], [0, 1, 1]]),
                           example_ids=list(range(8)),
                           feature_names=["a", "b", "c"],
                           labels={"my_label": ["yes", "no", "yes", "no", "yes", "no", "yes", "no"]},
                           encoding="random")

    label = Label(name="my_label", values=["yes", "no"], positive_class="yes")

    return enc_data, label

def check_params_recursively(params):
    if params is None or isinstance(params, (str, float, int)):
        return
    elif isinstance(params, list):
        for param in params:
            check_params_recursively(param)
    elif isinstance(params, dict):
        for key, value in params.items():
            check_params_recursively(key)
            check_params_recursively(value)
    else:
        raise ValueError(f"get_params() should return a yaml-friendly parameter representation, consisting only of lists, dicts and strings. Found: {params} of type {type(params)}")

def check_predictions(ml_method_instance, predictions, label, enc_data):
    error_mssg = f"The output of {ml_method_instance.__class__.__name__}.predict() is assumed to be a dictionary where the label name is the key, and its value is a numpy array with classes (one prediction per example)." \
                 "\nFor example: {" + f"'{label.name}': {[random.choice(label.values) for i in range(enc_data.examples.shape[0])]}" + "}" \
                                                                                                                                      f"\nFound the follwing instead: {predictions}"
    assert isinstance(predictions, dict), error_mssg + f"\n\nFound the wrong type: {type(predictions)}"
    assert list(predictions.keys()) == [label.name], error_mssg + f"\n\nFound the wrong keys: {list(predictions.keys())}"

    assert len(list(predictions["my_label"])) == enc_data.examples.shape[0], error_mssg + f"\n\nFound the wrong number of predictions."
    for pred in predictions["my_label"]:
        assert pred in label.values, error_mssg + f"\n\nFound the wrong predicted class: {pred}"

def check_predictions_proba(ml_method_instance, proba_predictions, label, enc_data):
    class1_proba = [round(random.uniform(0, 1), 1) for p in range(enc_data.examples.shape[0])]
    class2_proba = [round(1 - p, 2) for p in class1_proba]

    correct_example = {label.name: {label.values[0]: np.array(class1_proba),label.values[1]: np.array(class2_proba)}}

    error_mssg = f"The output of {ml_method_instance.__class__.__name__}.predict_proba() is assumed to be a dictionary where the label name is the key, " \
                 f"and its value is a dictionary containing the probabilities for each class." \
                 f"\n\nFor example (numbers are random examples):\n\t{correct_example}" \
                 f"\n\nFound the follwing instead: \n\t{proba_predictions}"

    assert isinstance(proba_predictions, dict), error_mssg + f"\n\nFound the wrong type: {type(proba_predictions)}"
    assert list(proba_predictions.keys()) == [label.name], error_mssg + f"\n\nFound the wrong keys: {list(proba_predictions.keys())}"

    assert isinstance(proba_predictions[label.name], dict), error_mssg + f"\n\nFound the wrong type where an inner dict was found: {type(proba_predictions[label.name])}"
    assert sorted(proba_predictions[label.name].keys()) == sorted(label.values), error_mssg + f"\n\nFound the wrong keys for the inner dict: {proba_predictions[label.name].keys()}"

    proba1 = proba_predictions[label.name][label.values[0]]
    proba2 = proba_predictions[label.name][label.values[1]]

    assert len(proba1) == enc_data.examples.shape[0], error_mssg + f"\n\nFound the wrong number of predictions: {proba1} (expected to be of length {enc_data.shape[0]})"
    assert len(proba2) == enc_data.examples.shape[0], error_mssg + f"\n\nFound the wrong number of predictions: {proba2} (expected to be of length {enc_data.shape[0]})"

    for pred1, pred2 in zip(proba1, proba2):
        assert isinstance(pred1, float), error_mssg + f"\n\nFound the following value of type {type(pred1)} where float was expected: {pred1}"
        assert isinstance(pred2, float), error_mssg + f"\n\nFound the following value of type {type(pred2)} where float was expected: {pred2}"

        assert 1 >= pred1 >= 0, error_mssg + f"\n\nExpected all predictions to range between [0, 1], found: {pred1}"
        assert 1 >= pred2 >= 0, error_mssg + f"\n\nExpected all predictions to range between [0, 1], found: {pred2}"

        assert 1.01 >= pred1+pred2 >= 0.99, error_mssg + f"\n\nProbabilties for both classes are expected to (approximately) sum to 1, found: {pred1} + {pred2} = {pred1 + pred2}"

def check_model_fitting_and_prediction(ml_method_instance, tmp_path):
    enc_data, label = get_example_encoded_data()

    logging.info(f"Attempting to call {ml_method_instance.__class__.__name__}.fit()...")
    ml_method_instance.fit(enc_data, label, optimization_metric="balanced_accuracy")
    logging.info(f"...Succeeded to calling {ml_method_instance.__class__.__name__}.fit()...")

    logging.info(f"Attempting to call {ml_method_instance.__class__.__name__}.predict()...")
    predictions = ml_method_instance.predict(enc_data, label)
    check_predictions(ml_method_instance, predictions, label, enc_data)
    logging.info(f"...Succeeded calling {ml_method_instance.__class__.__name__}.predict(). Predictions are:\n{predictions}")

    if ml_method_instance.can_predict_proba():
        logging.info("Attempting to predict ")
        logging.info(f"{ml_method_instance.__class__.__name__}.can_predict_proba() returns True, attempting to call {ml_method_instance.__class__.__name__}.predict_proba()...")
        proba_predictions = ml_method_instance.predict_proba(enc_data, label)
        check_predictions_proba(ml_method_instance, proba_predictions, label, enc_data)
        logging.info(f"...Succeeded calling {ml_method_instance.__class__.__name__}.predict_proba(). Predictions are:\n{proba_predictions}")
    else:
        logging.info(f"{ml_method_instance.__class__.__name__}.can_predict_proba() returns False, {ml_method_instance.__class__.__name__}.predict_proba() will be ignored")

    logging.info(f"Checking {ml_method_instance.__class__.__name__}.get_params()...")
    params = ml_method_instance.get_params()
    assert isinstance(params, dict), f"Expected get_params() to return a dictionary, found the following: {params}"
    logging.info(f"Checking formatting of the following params: {params}")
    check_params_recursively(params)

    with (tmp_path / "get_params.yaml").open("w") as file:
        yaml.dump(params, file)
    logging.info(f"...checking {ml_method_instance.__class__.__name__}.get_params() done.")

def check_store_load(ml_method_instance, tmp_path):
    logging.info(f"Attempting to store and load {ml_method_instance.__class__.__name__}...")

    logging.info(f"Storing to {tmp_path}...")
    ml_method_instance.store(tmp_path)
    logging.info(f"...Storing done")
    logging.info(f"Loading model from {tmp_path}...")

    loaded_model = ReflectionHandler.get_class_by_name(ml_method_instance.__class__.__name__, 'ml_methods/')()
    loaded_model.load(tmp_path)
    logging.info(f"...Loading done")

    logging.info(f"Testing if original and re-loaded model make the same predictions...")
    example_data, label = get_example_encoded_data()
    orig_preds = ml_method_instance.predict(example_data, label)
    loaded_preds = loaded_model.predict(example_data, label)

    assert (orig_preds["my_label"] == loaded_preds["my_label"]).all(), f"When using the same example data to call .predict() on both the original model " \
                                                                       f"and the model loaded after storing, different results were found: " \
                                                                       f"\nOriginal model predictions: {orig_preds}" \
                                                                       f"\nLoaded model predictions: {loaded_preds}." \
                                                                       f"\nThis indicates an issue with storing and loading the model."

    logging.info(f"...The original and re-loaded model make the same predictions.")

    logging.info(f"...Succeeded storing and loading.")

def check_ml_method_class(ml_method_class, default_params, tmp_path):
    assert issubclass(ml_method_class, MLMethod), f"Error: '{ml_method_class.__name__}' is not a subclass of MLMethod"

    logging.info(f"Attempting to create a new {ml_method_class.__name__} object (with default parameters if given)...")
    ml_method_instance = ml_method_class(**default_params)
    logging.info(f"...Successfully created an instance of {ml_method_instance.__class__.__name__}")

    check_methods(ml_method_instance)
    check_model_fitting_and_prediction(ml_method_instance, tmp_path)
    check_store_load(ml_method_instance, tmp_path)

def run_checks(parsed_args):
    logging.info(f"Testing new ML method file: '{parsed_args.ml_method_file}'")

    assert os.path.isfile(parsed_args.ml_method_file), f"Error: '{parsed_args.ml_method_file}' is not a file"
    ml_method_file_path = Path(parsed_args.ml_method_file)

    check_ml_method_file(ml_method_file_path)
    ml_method_class = get_ml_method_class(ml_method_file_path)

    default_params_filepath = get_default_params_filepath_from_name(ml_method_file_path)
    default_params = check_default_args(default_params_filepath, parsed_args.no_default_parameters, class_name=ml_method_file_path.stem)

    check_ml_method_class(ml_method_class, default_params, tmp_path=Path(parsed_args.tmp_path))

    logging.info("\n\nCongratulations! If the script ran this far without encountering any error, "
                 f"\nyou have successfully integrated a new MLMethod into immuneML. "
                 "\n\nTo finalise your ML method: "
                 "\n- Test run your MLMethod with an appropriate Encoder using the TrainMLModel instruction. See examples in the immuneML documentation at https://docs.immuneml.uio.no"
                 "\n- Please take a look if any warnings were thrown in this log file, and consider resolving those. These issues are not critical for the performance of your MLMethod, but may be necessary to resolve before merging the code into the main immuneML branch."
                 "\n- Add class documentation containing a general description of the encoding, its parameters, and an example YAML snippet. "
                 "\n- Add a unit test for the new ML method. "
                 "\n\nFor more details, see https://docs.immuneml.uio.no/latest/developer_docs/how_to_add_new_ML_method.html")

def main(args):
    parsed_args = parse_commandline_arguments(args)

    setup_logger(parsed_args.log_file)
    set_tmp_path(parsed_args.tmp_path)

    try:
        run_checks(parsed_args)
    except Exception as e:
        logging.error("\n\nA critical error occurred when testing the new MLMethod:\n")
        logging.exception(e)

        raise e

if __name__ == "__main__":
    main(sys.argv[1:])
